#include "../client/client.h"
#include <proto/cdplayer.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <libraries/cdplayer.h>

static qboolean cdValid = false;
static qboolean cdPlaying = false;
static qboolean cdWasPlaying = false;
static qboolean cdInitialised = false;
static qboolean cdEnabled = false;
static byte     cdPlayTrack;
static byte     cdMaxTrack;
static int      cdEmulatedStart;
static int      cdEmulatedLength;

struct CD_TOC cdTable;

struct IOStdReq	*CD_Request;
struct MsgPort	*CD_Port;

struct Library *CDPlayerBase;
struct ExecIFace *IExec;
struct CDPlayerIFace *ICDPlayer;

cvar_t *cd_nocd = 0;
extern cvar_t *cd_device;
extern cvar_t *cd_unit;
cvar_t *cd_shuffle;

FILE *cdfil = 0;

void CDAudio_Play(int track, qboolean looping)
{
	extern qboolean ogg_started;
  if (ogg_started)
  {
	  qboolean OGG_Open2(int number);
	  
	  OGG_Open2(track-1);
  }
  if (!cdEnabled) return;
  if (!cdInitialised) return;
  
  ICDPlayer->CDReadTOC (&cdTable, CD_Request);
  if (cdTable.cdc_NumTracks>0) cdValid = true;
  else cdValid = false;

  cdMaxTrack = cdTable.cdc_NumTracks;
  if (cdMaxTrack>100) cdMaxTrack = 100;

  if ((track < 1) || (track > cdMaxTrack))
  {
    CDAudio_Stop();
    return;
  }
  if (cdTable.cdc_Flags[track-1]==1)
  {
	  CDAudio_Stop();
	  return;
  }

  if (!cdValid)
  {
	return;
  }

  if(cdPlaying)
  {
    if(cdPlayTrack == track) 
	{
		return;
	}
    CDAudio_Stop();
  }
  
  cdEmulatedLength = (cdTable.cdc_Addr[track-1]/75)*1000;
  cdEmulatedStart = Sys_Milliseconds();

  cdPlayTrack = track;
  cdPlaying = true;
  cdWasPlaying = false;

  ICDPlayer->CDPlay (track, cdTable.cdc_NumTracks, CD_Request);
}

void CDSetVolume(float volume)
{
	if (!cdInitialised) return;
	
	int vol = volume*255;
	struct CD_Volume vs;
	vs.cdv_Chan0 = vol;
	vs.cdv_Chan1 = vol;
	vs.cdv_Chan2 = vol;
	vs.cdv_Chan3 = vol;
	
	ICDPlayer->CDSetVolume(&vs, CD_Request);
}

void CDAudio_Stop(void)
{
  if (!cdInitialised) return;
  if (!cdEnabled) return;
  if (!cdPlaying) return;

  ICDPlayer->CDStop(CD_Request);
  cdWasPlaying = false;
  cdPlaying = false;
}


void CDAudio_Resume(void)
{	
  if (!cdEnabled) return;
  if (!cdWasPlaying) return;
  if (!cdValid) return;
  if (!cdInitialised) return;

  ICDPlayer->CDResume(true, CD_Request);

  cdPlaying = true;
  cdPlaying = false;
}

void CDAudio_Pause(void)
{
    int err;
	
	if (cdInitialised==false) return;
	
    if (!cdEnabled)
	return;

    if (!cdPlaying)
	return;	

	ICDPlayer->CDResume(false, CD_Request);
    cdWasPlaying = cdPlaying;
    cdPlaying = false;
}


void CDAudio_Update(void)
{
	if (cdInitialised==false) return;
    if (cd_nocd && cd_nocd->modified)
    {
		cd_nocd->modified = 0;

		if (cd_nocd->value)
			CDAudio_Pause();
		else
			CDAudio_Resume();
    }
    if(Sys_Milliseconds() > (cdEmulatedStart + cdEmulatedLength))
    {
      cdPlaying = false;
	  cd_shuffle = Cvar_Get("cd_shuffle","0",CVAR_ARCHIVE);
	  if (cd_shuffle->value==1)
	  {
		    if (!cdEnabled) return;
  
			ICDPlayer->CDReadTOC (&cdTable, CD_Request);
			if (cdTable.cdc_NumTracks>0) cdValid = true;
			else cdValid = false;

			cdMaxTrack = cdTable.cdc_NumTracks;
		  if (cdPlayTrack+1<=cdMaxTrack)
			CDAudio_Play(cdPlayTrack+1, true);
		  else
		  {
			  for (int i=1;i<=cdMaxTrack;i++)
			  {
				  if (cdTable.cdc_Flags[i-1]==0)
				  {
					  CDAudio_Play(i, true);
					  return;
				  }
			  }
		  }
	  }
	  else
	  {
		CDAudio_Play(cdPlayTrack, true);
	  }
    }
}

int CDAudio_HardwareInit()
{
	
		char devname[2048];
		char cdunit[2048];
	
		cd_device = Cvar_Get("cd_device","diskimage.device",CVAR_ARCHIVE);
		cd_unit = Cvar_Get("cd_unit","0",CVAR_ARCHIVE);
		
		if (!cd_device) return 0;
		if (!cd_unit) return 0;
		
		strcpy(devname,cd_device->string);
		if (!stricmp(devname,"env_quake2"))
		{
			    if (0<(IDOS->GetVar("env:Quake2/cd_device",devname,255,0)))
				{
				}
		}
		strcpy(cdunit,cd_unit->string);
		if (!stricmp(cdunit,"env_quake2"))
		{
			    if (0<(IDOS->GetVar("env:Quake2/cd_unit",cdunit,255,0)))
				{
				}
		}			
				
		if (!cd_device) return 0;
		if (!stricmp(cd_device->string,"Off")) return 0;
		IExec = (struct ExecIFace *)(*(struct ExecBase **)4)->MainInterface;
		CDPlayerBase = IExec->OpenLibrary ("cdplayer.library",37L);
		if (!CDPlayerBase)
		{
			    Com_Printf("You need cdplayer.library to use CD Audio\n");
				return 0;
		}
		ICDPlayer = (struct CDPlayerIFace *)IExec->GetInterface(CDPlayerBase, "main", 1, NULL);	
		if (!ICDPlayer)
		{
				IExec->CloseLibrary(CDPlayerBase);
				CDPlayerBase = 0;
				Com_Printf("You need cdplayer.library to use CD Audio\n");
				return 0;
		}
		
		if (CD_Port = IExec->AllocSysObject(ASOT_PORT, NULL))
		{
			if (CD_Request= IExec->AllocSysObjectTags(ASOT_IOREQUEST, ASOIOR_Size,sizeof (struct IOStdReq), ASOIOR_ReplyPort, CD_Port, TAG_DONE))
			{
				if(IExec->OpenDevice (devname, atoi(cdunit), (struct IORequest *) CD_Request, 0) != -1)
				{
					ICDPlayer->CDReadTOC (&cdTable, CD_Request);
				}
			}
		}

		ICDPlayer->CDReadTOC (&cdTable, CD_Request);	
		
		return 1;
}

int CDAudio_Init(void)
{
  if (cdInitialised) return 0;	
  cd_nocd = Cvar_Get("cd_nocd", "0", CVAR_ARCHIVE);
  
  if (cd_nocd->value==1) return -1;

  if(CDAudio_HardwareInit())
  {	  
    cdEnabled=true;
	cdInitialised = true;
    Com_Printf("CD Audio Initialized\n");
	return 0;
  }  
  return 0;
}


void CDAudio_Shutdown(void)
{
  if(!cdInitialised) return;  
  CDAudio_Stop();
  
  ICDPlayer->CDStop (CD_Request);	
  if(CD_Request) IExec->CloseDevice ((struct IORequest *) CD_Request);
  if(CD_Request) { IExec->FreeSysObject(ASOT_IOREQUEST, CD_Request); CD_Request = NULL; }
  if(CD_Port) { IExec->FreeSysObject(ASOT_PORT, CD_Port); CD_Port = NULL; }  
  if (ICDPlayer) IExec->DropInterface((struct Interface*)ICDPlayer);
  if (CDPlayerBase) IExec->CloseLibrary (CDPlayerBase);
  CDPlayerBase = 0;
  ICDPlayer = 0;
  CD_Request = 0;
  CD_Port = 0;
  cdInitialised = false;
}
