/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/*
** GLA_IMP.C
**
** This file contains ALL Amiga specific stuff having to do with the
** OpenGL refresh.  When a port is being made the following functions
** must be implemented by the port:
**
** GLimp_EndFrame
** GLimp_Init
** GLimp_Shutdown
** GLimp_SwitchFullscreen
**
*/
#include <exec/exec.h>
#include <exec/memory.h>
#include <exec/execbase.h>
//#include <mgl/gl.h>
#include <intuition/intuition.h>
#include <graphics/gfx.h>
#include <graphics/blitattr.h>
#include <proto/intuition.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/picasso96api.h>
#include <proto/timer.h>

#include "../ref_gl/gl_local.h"

struct ExecIFace   *qglIExec;


struct Library        *IntuitionBase = NULL;
struct Library        *GfxBase       = NULL;
struct Library        *P96Base       = NULL;
struct Library        *MiniGLBase    = NULL;
struct Device         *TimerBase = NULL;

struct IntuitionIFace *IIntuition    = NULL;
struct GraphicsIFace  *IGraphics     = NULL;
struct P96IFace       *IP96          = NULL;
struct MiniGLIFace    *IMiniGL       = NULL;
struct GLContextIFace *IGL           = NULL;
struct TimerIFace     *ITimer        = NULL;

struct Screen         *g_pScreen     = NULL;
struct Window         *g_pWindow     = NULL;
struct BitMap         *g_pBuffer[2]  = { NULL, NULL };

struct MsgPort     *timer_mp = NULL;
struct TimeRequest *timer_rq = NULL;

int dispwidth, dispheight;

cvar_t *gl_lockmode;
//cvar_t *gl_texturecache;
//cvar_t *gl_drawparticles;
//cvar_t *gl_point_size;
//cvar_t *gl_skipparticles; //surgeon
//cvar_t *gl_mintriarea; //surgeon
//cvar_t *gl_subdivide_size; //surgeon
//cvar_t *gl_affinemodels; //surgeon
//cvar_t *gl_mtexbuffersize; //surgeon
cvar_t *gl_framestat;

//cvar_t *gl_guardband;

GLboolean old_context = GL_FALSE;

int gl_lmaps_drawn = 0;
int gl_bpolies_drawn = 0;
int dispwidth, dispheight;


extern cvar_t *vid_fullscreen;
extern cvar_t *vid_ref;

void CloseLibs(void)
{
	if( ITimer )        { qglIExec->DropInterface( (struct Interface *)ITimer ); ITimer = NULL; }
	if( TimerBase )     { qglIExec->CloseDevice( (struct IORequest *)timer_rq ); TimerBase = NULL; }
	if( timer_rq )      { qglIExec->FreeSysObject( ASOT_IOREQUEST, timer_rq ); timer_rq = NULL; }
	if( timer_mp )      { qglIExec->FreeSysObject( ASOT_PORT, timer_mp ); timer_mp = NULL; }
	if( IMiniGL )       { qglIExec->DropInterface( (struct Interface *)IMiniGL ); IMiniGL = NULL; }
	if( MiniGLBase )    { qglIExec->CloseLibrary( MiniGLBase ); MiniGLBase = NULL; }
	if( IP96 )          { qglIExec->DropInterface( (struct Interface *)IP96 ); IP96 = NULL; }
	if( P96Base )       { qglIExec->CloseLibrary( P96Base ); P96Base = NULL; }
	if( IGraphics )     { qglIExec->DropInterface( (struct Interface *)IGraphics ); IGraphics = NULL; }
	if( GfxBase )       { qglIExec->CloseLibrary( GfxBase ); GfxBase = NULL; }
	if( IIntuition )    { qglIExec->DropInterface( (struct Interface *)IIntuition ); IIntuition = NULL; }
	if( IntuitionBase ) { qglIExec->CloseLibrary( IntuitionBase ); IntuitionBase = NULL; }
	if( IMiniGL )       { qglIExec->DropInterface( (struct Interface *)IMiniGL ); IMiniGL = NULL; }
	if( MiniGLBase )    { qglIExec->CloseLibrary( MiniGLBase ); MiniGLBase = NULL; }
}

int OpenLibs(void)
{
	// Christ knows why this is necessary. But it just is.
	qglIExec = (struct ExecIFace *)(*(struct ExecBase **)4)->MainInterface;

	IntuitionBase = qglIExec->OpenLibrary( "intuition.library", 53 );
	GfxBase       = qglIExec->OpenLibrary( "graphics.library", 53 );
	P96Base       = qglIExec->OpenLibrary( "Picasso96API.library", 2 );
	MiniGLBase    = qglIExec->OpenLibrary( "minigl.library", 2 );
  
	if( !IntuitionBase || !GfxBase || !P96Base || !MiniGLBase )
	{
		Sys_Error("Unable to initialise GL");
		CloseLibs();
		return 0;
	}
  
	IIntuition = (struct IntuitionIFace *)qglIExec->GetInterface( IntuitionBase, "main", 1, NULL );
	IGraphics  = (struct GraphicsIFace *) qglIExec->GetInterface( GfxBase,       "main", 1, NULL );
	IP96       = (struct P96IFace *)      qglIExec->GetInterface( P96Base,       "main", 1, NULL );
	IMiniGL    = (struct MiniGLIFace *)   qglIExec->GetInterface( MiniGLBase,    "main", 1, NULL );

	timer_mp = (struct MsgPort *)qglIExec->AllocSysObjectTags( ASOT_PORT, TAG_DONE );
	timer_rq = (struct TimeRequest *)qglIExec->AllocSysObjectTags( ASOT_IOREQUEST, ASOIOR_Size, sizeof( struct TimeRequest ), ASOIOR_ReplyPort, timer_mp, TAG_DONE );
  
	if( !IIntuition || !IGraphics || !IP96 || !IMiniGL || !timer_mp || !timer_rq )
	{
		Sys_Error("Unable to initialise GL");
		CloseLibs();
		return 0;
	}

	if( qglIExec->OpenDevice( TIMERNAME, UNIT_MICROHZ, (struct IORequest *)timer_rq, 0 ) != 0)
	{
		Sys_Error("Unable to initialise GL");
		CloseLibs();
		return 0;
	}

	TimerBase = (struct Device *)timer_rq->Request.io_Device;
	ITimer = (struct TimerIFace *)qglIExec->GetInterface( (struct Library *)TimerBase, "main", 1, NULL );
	if( !ITimer )
	{
		Sys_Error("Unable to initialise GL");
		CloseLibs();
		return 0;
	}

	return 1; 
}


void GL_SetLockMode(char *mode)
{
	if (0 == stricmp(mode, "SMART"))
	{
		Com_Printf("MiniGL Lock Mode: Smart\n");
		mglLockMode(MGL_LOCK_SMART);
	}
/*
	else if (0 == stricmp(mode, "AUTO"))
	{
		Com_Printf("MiniGL Lock Mode: Automatic\n");
		mglLockMode(MGL_LOCK_AUTOMATIC);
	}
*/
	else
	{
		Com_Printf("MiniGL Lock Mode: Manual (per Frame)\n");
		mglLockMode(MGL_LOCK_MANUAL);
	}
}


void Destroy_GL_Window( void )
{
	if( g_pBuffer[0] )
	{
		IP96->p96FreeBitMap( g_pBuffer[0] );
		g_pBuffer[0] = NULL;
	}

	if( g_pBuffer[1] )
	{
		IP96->p96FreeBitMap( g_pBuffer[1] );
		g_pBuffer[1] = NULL;
	}

	if( IGL )
	{
		IGL->DeleteContext();
		IGL = NULL;
	}

	if( g_pWindow )
	{
		IIntuition->CloseWindow( g_pWindow );
		g_pWindow = NULL;
	}

	if( g_pScreen )
	{
		IIntuition->CloseScreen( g_pScreen );
		g_pScreen = NULL;
	}
}

/*
** VID_CreateWindow
*/
#define WINDOW_CLASS_NAME       "Quake 2"

qboolean VID_CreateWindow( int width, int height, qboolean fullscreen )
{
	cvar_t                  *vid_xpos, *vid_ypos;
	int                     x, y, w, h;
	int                     exstyle;
	cvar_t                  *depth, *buffers, *cv;
	ULONG   wflgs;
	Tag     screentag;
	Tag     titletag;

	Destroy_GL_Window();

	depth = ri.Cvar_Get( "gl_forcedepth", "16", CVAR_ARCHIVE );
	gl_framestat = ri.Cvar_Get( "gl_framestat", "0", 0 );

	// Try and open a screen
	if (fullscreen)
	{
		uint32 ModeID, errorCode;

		x = 0;
		y = 0;

		ModeID = IP96->p96BestModeIDTags( P96BIDTAG_NominalWidth,  width,
		                                  P96BIDTAG_NominalHeight, height,
		                                  P96BIDTAG_Depth,         16, //depth ? depth->value : 16,
		                                  TAG_DONE );
		

		g_pScreen = IIntuition->OpenScreenTags(NULL,
			SA_DisplayID,           ModeID,
			SA_Width,               width,
			SA_Height,              height,
			SA_Depth,               16, //depth ? depth->value : 16,
			SA_ShowTitle,           FALSE,
			SA_Quiet,               TRUE,
			SA_FullPalette,         TRUE,
			SA_ErrorCode,           (ULONG)&errorCode,
			TAG_DONE );

		if( !g_pScreen )
		{
			ri.Con_Printf(PRINT_ALL, "OpenScreen: ");
			switch(errorCode)
			{
				case OSERR_NOMONITOR:
					ri.Con_Printf(PRINT_ALL, "Monitor for display mode not available\n");
					break;
				case OSERR_NOCHIPS:
					ri.Con_Printf(PRINT_ALL, "Custom chips for display mode missing\n");
					break;
				case OSERR_NOMEM:
					ri.Con_Printf(PRINT_ALL, "Out of system memory\n");
					break;
				case OSERR_NOCHIPMEM:
					ri.Con_Printf(PRINT_ALL, "Out of chip memory\n");
					break;
				case OSERR_PUBNOTUNIQUE:
					ri.Con_Printf(PRINT_ALL, "Public screen name already in use\n");
					break;
				case OSERR_UNKNOWNMODE:
					ri.Con_Printf(PRINT_ALL, "Didn't recognize display mode requested\n");
					break;
				case OSERR_TOODEEP:
					ri.Con_Printf(PRINT_ALL, "This hardware cannot display screens of that depth\n");
					break;
				case OSERR_ATTACHFAIL:
					ri.Con_Printf(PRINT_ALL, "An illegal attachment of screens was requested\n");
					break;
				default:
					ri.Con_Printf(PRINT_ALL, "Error code %s\n", errorCode);
					break;
			}
			return false;
		}		
	}

	// If there is no screen, try and open on WB
	if( !g_pScreen )
	{
		fullscreen = false;

		vid_xpos = ri.Cvar_Get ("vid_xpos", "0", 0);
		vid_ypos = ri.Cvar_Get ("vid_ypos", "0", 0);
		x = vid_xpos ? vid_xpos->value : 0;
		y = vid_ypos ? vid_ypos->value : 0;

		screentag = TAG_IGNORE;
		titletag = WA_Title;
		wflgs = WFLG_REPORTMOUSE
			| WFLG_NOCAREREFRESH
			| WFLG_ACTIVATE
			| WFLG_RMBTRAP
			| WFLG_DRAGBAR
			| WFLG_DEPTHGADGET
			| WFLG_CLOSEGADGET;
	}
	else
	{
			screentag = WA_CustomScreen;
			titletag = TAG_IGNORE;
			wflgs = WFLG_BORDERLESS
				| WFLG_REPORTMOUSE
				| WFLG_NOCAREREFRESH
				| WFLG_ACTIVATE
				| WFLG_RMBTRAP;
	}

	g_pWindow = IIntuition->OpenWindowTags(NULL,
		WA_Left,            (ULONG)x,
		WA_Top,             (ULONG)y,
		WA_Flags,           wflgs,
		WA_InnerWidth,      width,
		WA_InnerHeight,     height,
		screentag,          (ULONG)g_pScreen,
		titletag,           (ULONG)"Quake 2",
		WA_ScreenTitle,     (ULONG)"Quake 2 - Help / Scroll Lock captures mouse",
		TAG_DONE);

	if( !g_pWindow )
	{
		Destroy_GL_Window();
		ri.Sys_Error( ERR_FATAL, "Couldn't create window" );
		return false;
	}

	g_pBuffer[0] = IP96->p96AllocBitMap( width, height, 16 /*depth ? depth->value : 16*/, BMF_MINPLANES | BMF_DISPLAYABLE, g_pWindow->RPort->BitMap, 0 );
	if( !g_pBuffer[0] )
	{
		Destroy_GL_Window();
		ri.Sys_Error( ERR_FATAL, "Couldn't allocate a bitmap" );
		return false;
	}

	g_pBuffer[1] = IP96->p96AllocBitMap( width, height, 16 /*depth ? depth->value : 16*/, BMF_MINPLANES | BMF_DISPLAYABLE, g_pWindow->RPort->BitMap, 0 );
	if( !g_pBuffer[1] )
	{
		Destroy_GL_Window();
		ri.Sys_Error( ERR_FATAL, "Couldn't allocate a bitmap" );
		return false;
	}

	IGL = IMiniGL->CreateContextTags(
		MGLCC_PrivateBuffers,   2,
		MGLCC_FrontBuffer,      g_pBuffer[0],
		MGLCC_BackBuffer,       g_pBuffer[1],
		MGLCC_PixelDepth,       16, //depth ? depth->value : 16,
		MGLCC_StencilBuffer,    TRUE,
		MGLCC_VertexBufferSize, 1<<17,
		TAG_DONE );
	if( !IGL )
	{
		Destroy_GL_Window();
		ri.Sys_Error( ERR_FATAL, "Couldn't create GL context" );
		printf( "GL context epic fail :-(\n" );
		return false;
	}

	IGL->GLViewport( 0, 0, width, height );
	mglMakeCurrent( IGL );
	
	gl_lockmode = ri.Cvar_Get( "gl_lockmode", "SMART", CVAR_ARCHIVE );
	GL_SetLockMode( gl_lockmode ? gl_lockmode->string : "SMART" );

	dispwidth = width;
	dispheight = height;

	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow( width, height );

	return true;
}


/*
** GLimp_SetMode
*/
int GLimp_SetMode( int *pwidth, int *pheight, int mode, qboolean fullscreen )
{
	int width, height;
	const char *win_fs[] = { "W", "FS" };

	ri.Con_Printf( PRINT_ALL, "Initializing OpenGL display\n");
	ri.Con_Printf (PRINT_ALL, "...setting mode %d:", mode );

	if ( !ri.Vid_GetModeInfo( &width, &height, mode ) )
	{
		ri.Con_Printf( PRINT_ALL, " invalid mode\n" );
		return rserr_invalid_mode;
	}

	// do a CDS if needed
	if (fullscreen)
	{
		ri.Con_Printf( PRINT_ALL, "fullscreen\n" );
		*pwidth = width;
		*pheight = height;
		if ( !VID_CreateWindow (width, height, true) )
			return rserr_invalid_mode;
	}
	else
	{
		ri.Con_Printf( PRINT_ALL, "windowed\n" );
		*pwidth = width;
		*pheight = height;
		if ( !VID_CreateWindow (width, height, false) )
			return rserr_invalid_mode;
	}

	return rserr_ok;
}

/*
** GLimp_Shutdown
**
** This routine does all OS specific shutdown procedures for the OpenGL
** subsystem.  Under OpenGL this means NULLing out the current DC and
** HGLRC, deleting the rendering context, and releasing the DC acquired
** for the window.  The state structure is also nulled out.
**
*/
void GLimp_Shutdown( void )
{
	Destroy_GL_Window();
	CloseLibs();
}

/*
void GLimp_TextureSwap(void *data, int ordinal)
{
        image_t *img = (image_t *)data;
        img->is_swapped = true;
        glDeleteTextures(1, &img->texnum);
}
*/

/*
void GLimp_TextureReload(void *data, int ordinal)
{
        extern void GL_RestoreImage(image_t *image);

        image_t *img = (image_t *)data;
        GL_RestoreImage(img);
        img->is_swapped = false;
}
*/

/*
** GLimp_Init
**
** This routine is responsible for initializing the OS specific portions
** of OpenGL.  Under Win32 this means dealing with the pixelformats and
** doing the wgl interface stuff.
*/

int GLimp_Init( void *hinstance, void *wndproc )
{
	if( !OpenLibs() ) return 0;
	return 1;
}


/*
** GLimp_BeginFrame
*/
void GLimp_BeginFrame( float camera_separation )
{
	extern cvar_t *gl_debug;
	static int timed = 0;
	static struct TimeVal tv;
	ULONG ms;

  //kprintf("Stat: %i %i\n",current,peak);

	if (timed == 0)
	{
		ITimer->GetSysTime(&tv);
		timed = 1;
	}
	else
	{
		struct TimeVal end;
		struct TimeVal start;
	    
		start = tv;

		ITimer->GetSysTime(&end);
		tv = end;
          
		ITimer->SubTime(&end,&start);  
		ms = end.Microseconds;
	}

	if( gl_lockmode && gl_lockmode->modified )
	{
		GL_SetLockMode(gl_lockmode->string);
		gl_lockmode->modified = false;
	}


	glHint(MGL_W_ONE_HINT, GL_FASTEST);

	mglLockDisplay();

	if( gl_framestat && gl_framestat->value == 1 )
	{
	    ri.Con_Printf(PRINT_ALL, "Frame time: %d microseconds\n", ms);
	    ri.Con_Printf(PRINT_ALL, "FPS: %3.3f fps\n", 1000000.0/(float)ms);
	    ri.Con_Printf(PRINT_ALL, "LM: %d   BP: %d\n", gl_lmaps_drawn, gl_bpolies_drawn);
	}

	gl_lmaps_drawn = 0;
	gl_bpolies_drawn = 0;
}

/*
** GLimp_EndFrame
** 
** Responsible for doing a swapbuffers and possibly for other stuff
** as yet to be determined.  Probably better not to make this a GLimp
** function and instead do a call to GLimp_SwapBuffers.
*/
void GLimp_EndFrame (void)
{
	GLint bufn;
	struct BitMap *buf;
	

	if( !IGL ) return;

	mglUnlockDisplay();
	IGL->MGLWaitGL();

	glGetIntegerv( GL_DRAW_BUFFER, &bufn );
	
	buf = NULL;
	if( bufn == GL_BACK  ) buf = g_pBuffer[1];
	if( bufn == GL_FRONT ) buf = g_pBuffer[0];
	if( buf ) IGraphics->BltBitMapRastPort( buf, 0, 0, g_pWindow->RPort, g_pWindow->BorderLeft, g_pWindow->BorderTop, dispwidth, dispheight, 0x0c0 );

	IGraphics->BltBitMapTags(
		BLITA_Source,   g_pBuffer[1],
		BLITA_SrcType,  BLITT_BITMAP,
		BLITA_SrcX,     0,
		BLITA_SrcY,     0,
		BLITA_Dest,     g_pBuffer[0],
		BLITA_DestType, BLITT_BITMAP,
		BLITA_DestX,    0,
		BLITA_DestY,    0,
		BLITA_Width,    dispwidth,
		BLITA_Height,   dispheight,
		BLITA_Minterm,  0x0c0,
		TAG_DONE );

	buf = g_pBuffer[0];
	g_pBuffer[0] = g_pBuffer[1];
	g_pBuffer[1] = buf;

	IGL->MGLUpdateContextTags( MGLCC_FrontBuffer, g_pBuffer[0], MGLCC_BackBuffer, g_pBuffer[1], TAG_DONE );
}

/*
** GLimp_AppActivate
*/
void GLimp_AppActivate( qboolean active )
{
}


struct Window *GetWindowHandle(void)
{
    return g_pWindow;
}

