#include "client.h"
#include "in.h"

#include <amigainput/amigainput.h>

#include <proto/amigainput.h>
#include <proto/exec.h>

struct Library   *Q2_AIN_Base;
struct AIN_IFace *Q2_IAIN;

cvar_t  *joy_name;
cvar_t  *joy_advanced;
cvar_t  *joy_advaxisx;
cvar_t  *joy_advaxisy;
cvar_t  *joy_advaxisz;
cvar_t  *joy_advaxisr;
cvar_t  *joy_advaxisu;
cvar_t  *joy_advaxisv;
cvar_t  *joy_forwardthreshold;
cvar_t  *joy_sidethreshold;
cvar_t  *joy_upthreshold;
cvar_t  *joy_pitchthreshold;
cvar_t  *joy_yawthreshold;
cvar_t  *joy_pitchsensitivity;
cvar_t  *joy_yawsensitivity;

cvar_t *joy_enablehats;

cvar_t  *joy_side_reverse;
cvar_t  *joy_up_reverse;
cvar_t  *joy_forward_reverse;

static APTR joystick_context;
char joystick_name[1024];
int joystick_id = -1;
int joystick_num_axes    = 0;
int joystick_num_buttons = 0;
int joystick_num_hats    = 0;

#define JOYSTICK_OFFSET(buffer, offset)	(((int32 *)buffer)[offset])

struct ain_joystick
{
	AIN_DeviceID	 id;
	const char      *name;
};

struct ain_data
{
	AIN_DeviceHandle   *handle;
	APTR				context;

	uint32				axisBufferOffset[8];
	int32				axisData[8];
	char                axisName[8][32];

	uint32				buttonBufferOffset[16];
	int32				buttonData[16];

	uint32				hatBufferOffset[2];
	int32				hatData[2];
};

struct joystickData
{
	APTR             context;
	int          *count;
	struct ain_joystick *list;
};

int joystick_count = 0;
struct ain_joystick joystick_list[8];
struct ain_data *joystick_data = 0;

static char *Q2_strdup(const char *str)
{
	char *duplicate = NULL;

	if (str)
	{
		int len = strlen(str);

		if((duplicate = IExec->AllocVecTags(len + 1, TAG_DONE))) 
		{
			memcpy(duplicate, str, len + 1);
		}
	}
	return duplicate;
}

static void Q2_freestr(const char *str)
{
	if (str)
	{
		IExec->FreeVec((char*)str);
	}
}

BOOL enumeration_function(AIN_Device *device, void *UserData)
{
	APTR context =  ((struct joystickData *)UserData)->context;
	int *count   =  ((struct joystickData *)UserData)->count;
	struct ain_joystick *ain_joy = &((struct joystickData *)UserData)->list[*count];

	BOOL result = FALSE;

	if (*count < 8)
	{
		if (device->Type == AINDT_JOYSTICK)
		{
			if ((device->NumAxes > 0) && (device->NumButtons > 0))
			{
				AIN_DeviceHandle *handle = Q2_IAIN->AIN_ObtainDevice (context, device->DeviceID);

				if (handle)
				{
					ain_joy->id   = device->DeviceID;
					ain_joy->name = Q2_strdup(device->DeviceName);
					if (!stricmp(ain_joy->name,joystick_name))
					{
						joystick_id = ain_joy->id;
					}
					

					(*count)++;

					Q2_IAIN->AIN_ReleaseDevice (context, handle);

					result = TRUE;
				}
			}
		}
    }
	return result;
}

BOOL MyInitJoy()
{
	Q2_AIN_Base = IExec->OpenLibrary("AmigaInput.library",51);
	if (Q2_AIN_Base)
	{
		Q2_IAIN = (struct AIN_IFace *)IExec->GetInterface(Q2_AIN_Base,"main",1,NULL);
		if(!Q2_IAIN)
		{	   
			return -1;
		}
	}
	
	joystick_context = Q2_IAIN->AIN_CreateContext(1, NULL);   
   
	if (joystick_context)
	{
		joystick_count = 0;
		struct joystickData data = {
				 joystick_context,
				&joystick_count,
				&joystick_list[0]
		};

		Q2_IAIN->AIN_EnumDevices(joystick_context, enumeration_function, &data);
		if (joystick_count>0)
		{
			if (joystick_id == -1)
			{
				joystick_id = joystick_list[0].id;
			}
		}
	}
	else
	{
		return -1;
	}		
   
   return (joystick_count>0);
}

void IN_StartupJoystick (void) 
{ 
    BOOL result;
    cvar_t *cv;
	APTR handle;
	char tmpstr[32];
	int tmpoffset;
	int i;
	
	strcpy(joystick_name,"");
	joystick_id = -1;
	
    cv = Cvar_Get ("joy_name", "joystick", CVAR_ARCHIVE);	
	strcpy(joystick_name,cv->string);

	result = MyInitJoy();
    if (result == -1)
    {
			if (Q2_AIN_Base)    IExec->DropInterface((struct Interface *)Q2_IAIN);
			if (Q2_AIN_Base) IExec->CloseLibrary(Q2_AIN_Base);
			Q2_IAIN = NULL;
			Q2_AIN_Base = NULL;
			return;
    }
	handle = Q2_IAIN->AIN_ObtainDevice(joystick_context, joystick_id);
	if (!handle)
	{
			if (Q2_AIN_Base)    IExec->DropInterface((struct Interface *)Q2_IAIN);
			if (Q2_AIN_Base) IExec->CloseLibrary(Q2_AIN_Base);
			Q2_IAIN = NULL;
			Q2_AIN_Base = NULL;
			return;		
	}
	
	joystick_data = IExec->AllocVecTags(sizeof (struct ain_data), AVT_ClearWithValue, 0, TAG_DONE );

	if (joystick_data)
	{	
		joystick_data->handle = handle;
		joystick_data->context = joystick_context;
		
		result = TRUE;
		
		result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_NUMAXES,    0, &joystick_num_axes, 4);
		result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_NUMBUTTONS, 0, &joystick_num_buttons, 4);
		result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_NUMHATS,    0, &joystick_num_hats, 4);
		
		joystick_num_axes = joystick_num_axes <= 8 ? joystick_num_axes : 8;
		joystick_num_buttons = joystick_num_buttons <= 16 ? joystick_num_buttons : 16;
		joystick_num_hats = joystick_num_hats <= 2 ? joystick_num_hats : 2;
		
		for (i = 0; i < 8; i++)
		{
			  joystick_data->axisName[i][0] = 0;
		}
		
		for (i = 0; i < joystick_num_axes; i++)
		{
			result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_AXIS_OFFSET, i, &(joystick_data->axisBufferOffset[i]), 4);
			result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_AXISNAME,    i, &(joystick_data->axisName[i][0]), 32 );
		}
		
		for (i = 0; i < joystick_num_axes; i++)
		{
			if (i!=0)
			{
			  if (!strcasecmp(&joystick_data->axisName[i][0], "X-Axis" ))
			  {
			    tmpoffset = joystick_data->axisBufferOffset[0];
			    strncpy(tmpstr, joystick_data->axisName[0], 32);
			    
			    joystick_data->axisBufferOffset[0] = joystick_data->axisBufferOffset[i];
			    strncpy(joystick_data->axisName[0], joystick_data->axisName[i], 32);
			    
			    joystick_data->axisBufferOffset[i] = tmpoffset;
			    strncpy(joystick_data->axisName[i], tmpstr, 32);
			    
			    continue;
			  }
			}

			if (i!=1)
			{
			  if (!strcasecmp(&joystick_data->axisName[i][0], "Y-Axis" ))
			  {
			    tmpoffset = joystick_data->axisBufferOffset[1];
			    strncpy(tmpstr, joystick_data->axisName[1], 32);
			    
			    joystick_data->axisBufferOffset[1] = joystick_data->axisBufferOffset[i];
			    strncpy(joystick_data->axisName[1], joystick_data->axisName[i], 32);
			    
			    joystick_data->axisBufferOffset[i] = tmpoffset;
			    strncpy(joystick_data->axisName[i], tmpstr, 32);
			    
			    continue;
			  }
			}
		}

		for (i = 0; i < joystick_num_buttons; i++)
		{
			result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_BUTTON_OFFSET, i, &(joystick_data->buttonBufferOffset[i]), 4);
		}

		for (i = 0; i < joystick_num_hats; i++)
		{
			result = result && Q2_IAIN->AIN_Query(joystick_data->context, joystick_id, AINQ_HAT_OFFSET, i, &(joystick_data->hatBufferOffset[i]), 4);
		}		
		
		if (result)
		{
			return;
		}
	}
	
	if (!result)
	{
			Q2_IAIN->AIN_ReleaseDevice (joystick_context, handle);
			for (int i=0;i<joystick_count;i++)
			{
				if (joystick_list[i].name!=0) 
				{
					Q2_freestr(joystick_list[i].name);
				}
				joystick_list[i].name = 0;
			}
			if (joystick_data) IExec->FreeVec(joystick_data);
			joystick_data = 0;
			if (Q2_IAIN) IExec->DropInterface(Q2_IAIN);			
			if (Q2_AIN_Base) IExec->CloseLibrary(Q2_AIN_Base);
			Q2_IAIN = NULL;
			Q2_AIN_Base = NULL;
			return;
	}
}

void IN_ShutdownJoystick(void)
{
  cvar_t *cv;
  
  if (!Q2_IAIN) return;	
	
  Q2_IAIN->AIN_ReleaseDevice (joystick_context, joystick_data->handle);
  for (int i=0;i<joystick_count;i++)
  {
	if (joystick_list[i].name!=0) 
	{
		Q2_freestr(joystick_list[i].name);
	}
	joystick_list[i].name = 0;
  }
  if (joystick_data) IExec->FreeVec(joystick_data);
  joystick_data = 0;
  if (Q2_IAIN)    IExec->DropInterface((struct Interface *)Q2_IAIN);
  if (Q2_AIN_Base) IExec->CloseLibrary(Q2_AIN_Base);
  Q2_IAIN = NULL;
  Q2_AIN_Base = NULL;
}

float GetStickMovement(float value, float speed, cvar_t *threshhold, cvar_t *reverse)
{
    float   result = 0.0;

    if(fabs(value) > threshhold->value)
    {
			if(value > 0)
			{
		    result = speed * 64;
			}
			else
			{
		    result = -speed * 64;
			}
    }
    if(reverse->value)
    {
			result = -result;
    }
    return(result);
}

cvar_t *usehats = 0;
cvar_t *hatleft = 0;
cvar_t *hatright = 0;
cvar_t *hattop = 0;
cvar_t *hatbottom = 0;

void IN_JoyMove (usercmd_t *cmd)
{
    float   speed, aspeed;
    int   fAxisValue;
    int joy_key;
	void *buffer;
	int i;
	int hattop_value=8, hatbottom_value=5, hatleft_value=3, hatright_value=7;


	usehats = Cvar_Get ("in_enablehats", "1", CVAR_ARCHIVE);
	hattop = Cvar_Get("in_hattop","8",CVAR_ARCHIVE);
	hatbottom = Cvar_Get("in_hatbottom","5",CVAR_ARCHIVE);
	hatright = Cvar_Get("in_hatleft","3",CVAR_ARCHIVE);
	hatleft = Cvar_Get("in_hatright","7",CVAR_ARCHIVE);
	if (hattop) hattop_value = hattop->value;
	if (hatbottom) hatbottom_value = hatbottom->value;
	if (hatleft) hatleft_value = hatleft->value;
	if (hatright) hatright_value = hatright->value;

    if (!Q2_IAIN) return;

    if ( (in_speed.state & 1) ^ (int)cl_run->value) speed = 2;
    else speed = 1;

    aspeed = speed * cls.frametime;

	if (joystick_data && Q2_IAIN->AIN_ReadDevice(joystick_data->context, joystick_data->handle, &buffer))
	{
		for (i = 0; i < joystick_num_axes; i++)
		{
			int axisdata = JOYSTICK_OFFSET(buffer, joystick_data->axisBufferOffset[i]);

			if (axisdata > 32767)  axisdata =  32767;
			if (axisdata < -32768) axisdata = -32768;
			
			fAxisValue = axisdata;
			    
			fAxisValue -= 32768;

			fAxisValue /= 32768;		

			if(axisdata != joystick_data->axisData[i]) 
			{	
				joystick_data->axisData[i] = axisdata;
				

				{
					if (axisdata<-30000)
					{
						Key_Event(K_AUX13+i*2,false,cls.realtime,-1);
						Key_Event(K_AUX13+i*2,true,cls.realtime,-1);
					}
					else if (axisdata>30000)
					{
						Key_Event(K_AUX14+i*2,false,cls.realtime,-1);
						Key_Event(K_AUX14+i*2,true,cls.realtime,-1);
					}
					else
					{
						Key_Event(K_AUX13+i*2,false,cls.realtime,-1);
						Key_Event(K_AUX14+i*2,false,cls.realtime,-1);
					}
				}
			}
		}
		
		if (usehats->value)
		{
			for (i = 0; i < joystick_num_hats; i++)
			{
				int hatdata = JOYSTICK_OFFSET(buffer, joystick_data->hatBufferOffset[i]);
			
				if (hatdata != joystick_data->hatData[i])
				{				
					if ((i==0)&&(hatdata>0))
					{
						if (hatdata==hattop_value) joy_key = K_AUX25;
						else if (hatdata == hatleft_value) joy_key = K_AUX26;
						else if (hatdata == hatbottom_value) joy_key = K_AUX27;
						else if (hatdata == hatright_value) joy_key = K_AUX28;
						else joy_key = 0;
					
						Key_Event(K_AUX25,false,cls.realtime,-1);
						Key_Event(K_AUX26,false,cls.realtime,-1);
						Key_Event(K_AUX27,false,cls.realtime,-1);
						Key_Event(K_AUX28,false,cls.realtime,-1);
					
						if (joy_key>0)
						{
							if (hatdata)
							{
								Key_Event(joy_key,false,cls.realtime,-1);
								Key_Event(joy_key,true,cls.realtime,-1);
							}
							else
							{
								Key_Event(joy_key,false,cls.realtime,-1);
							}
						}
					}
					else if ((i==1)&&(hatdata>0))
					{		
						Key_Event(K_AUX29,false,cls.realtime,-1);
						Key_Event(K_AUX30,false,cls.realtime,-1);
						Key_Event(K_AUX31,false,cls.realtime,-1);
						Key_Event(K_AUX32,false,cls.realtime,-1);				
				
						if (hatdata == hattop_value) joy_key = K_AUX29;
						else if (hatdata == hatleft_value) joy_key = K_AUX30;
						else if (hatdata == hatbottom_value) joy_key = K_AUX31;
						else if (hatdata == hatright_value) joy_key = K_AUX32;
						else joy_key = 0;
						if (joy_key>0)
						{
							if (hatdata)
							{
								Key_Event(joy_key,true,cls.realtime,-1);
							}
							else
							{
								Key_Event(joy_key,false,cls.realtime,-1);
							}	
						}					
					}
				}
				else
				{
						Key_Event(K_AUX25,false,cls.realtime,-1);
						Key_Event(K_AUX26,false,cls.realtime,-1);
						Key_Event(K_AUX27,false,cls.realtime,-1);
						Key_Event(K_AUX28,false,cls.realtime,-1);
				}				
			}
		}
		
		for (i = 0; i < joystick_num_buttons; i++)
		{
			int buttondata = JOYSTICK_OFFSET(buffer, joystick_data->buttonBufferOffset[i]);

			if(buttondata != joystick_data->buttonData[i]) 
			{		
				if (i==0)
				{						
					joy_key = K_JOY1;
					if (buttondata)
					{
						Key_Event(joy_key+3,false,cls.realtime,-1);
						Key_Event(joy_key+3,true,cls.realtime,-1);
					}
					else
					{
						Key_Event(joy_key+3,false,cls.realtime,-1);		
					}
				}
				else if (i==1)
				{
					joy_key = K_JOY1;
					if (buttondata)
					{
						Key_Event(joy_key,false,cls.realtime,-1);						
						Key_Event(joy_key,true,cls.realtime,-1);
					}
					else
					{
						Key_Event(joy_key,false,cls.realtime,-1);		
					}					
				}
				else if (i==2)
				{
					joy_key = K_JOY1;
					if (buttondata)
					{
						Key_Event(joy_key+2,false,cls.realtime,-1);						
						Key_Event(joy_key+2,true,cls.realtime,-1);
					}
					else
					{
						Key_Event(joy_key+2,false,cls.realtime,-1);		
					}					
				}	
				else if (i==3)
				{
					joy_key = K_JOY1;
					if (buttondata)
					{
						Key_Event(joy_key+1,false,cls.realtime,-1);						
						Key_Event(joy_key+1,true,cls.realtime,-1);
					}
					else
					{
						Key_Event(joy_key+1,false,cls.realtime,-1);		
					}					
				}				
				else if ((i>=4)&&(i<=28))
				{
					joy_key = K_AUX1;
					if (buttondata)
					{
						Key_Event(joy_key+(i-4),false,cls.realtime,-1);						
						Key_Event(joy_key+(i-4),true,cls.realtime,-1);
					}
					else
					{
						Key_Event(joy_key+(i-4),false,cls.realtime,-1);
					}
				}
				joystick_data->buttonData[i] = buttondata;
			}
			
		}		
	}
}

void IN_Commands (void)
{
}

void IN_InitJoy()
{
    joy_name                = Cvar_Get ("joy_name",                 "joystick", 0);
    joy_advanced            = Cvar_Get ("joy_advanced",             "0",        0);
    joy_advaxisx            = Cvar_Get ("joy_advaxisx",             "0",        0);
    joy_advaxisy            = Cvar_Get ("joy_advaxisy",             "0",        0);
    joy_advaxisz            = Cvar_Get ("joy_advaxisz",             "0",        0);
    joy_advaxisr            = Cvar_Get ("joy_advaxisr",             "0",        0);
    joy_advaxisu            = Cvar_Get ("joy_advaxisu",             "0",        0);
    joy_advaxisv            = Cvar_Get ("joy_advaxisv",             "0",        0);
    joy_forwardthreshold    = Cvar_Get ("joy_forwardthreshold",     "0.15",     0);
    joy_sidethreshold       = Cvar_Get ("joy_sidethreshold",        "0.15",     0);
    joy_upthreshold         = Cvar_Get ("joy_upthreshold",          "0.15",     0);
    joy_pitchthreshold      = Cvar_Get ("joy_pitchthreshold",       "0.15",     0);
    joy_yawthreshold        = Cvar_Get ("joy_yawthreshold",         "0.15",     0);
    joy_pitchsensitivity    = Cvar_Get ("joy_pitchsensitivity",     "1",        0);
    joy_yawsensitivity      = Cvar_Get ("joy_yawsensitivity",       "-1",       0);
    joy_up_reverse          = Cvar_Get ("joy_up_reverse",           "0",        CVAR_ARCHIVE);
    joy_forward_reverse     = Cvar_Get ("joy_forward_reverse",      "0",        CVAR_ARCHIVE);
    joy_side_reverse        = Cvar_Get ("joy_side_reverse",         "0",        CVAR_ARCHIVE);
	joy_enablehats			= Cvar_Get ("joy_enablehats",           "0",        CVAR_NOSET);

}
