/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/*
** RW_IMP.C
**
** This file contains ALL Amiga specific stuff having to do with the
** software refresh.  When a port is being made the following functions
** must be implemented by the port:
**
** SWimp_EndFrame
** SWimp_Init
** SWimp_SetPalette
** SWimp_Shutdown
*/
#include "../ref_soft/r_local.h"
#include <intuition/intuition.h>
#include <intuition/screens.h>
#include <graphics/gfx.h>
#include <exec/exec.h>

#include <proto/exec.h>
//#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/picasso96api.h>

#include "sw_imp.h"

//#include "dll.h"

struct sodata sod;

static qboolean libs_init = false;
static qboolean screen_public = true;

static struct Window *g_pWindow;
static struct Screen *g_pScreen;
static struct BitMap *g_pFriend;

static uint16 word_pal[256];
static uint32 long_pal[256];
unsigned char first_palette[4*256];

#define MAYBE_OPEN_LIBS if (libs_init == false) SWimp_OpenLibs()

qboolean SWimp_OpenLibs(void)
{
    if (libs_init) return;
    libs_init = true;

    sod.P96Base = sod.IExec->OpenLibrary( "Picasso96API.library", 2L );
    if( !sod.P96Base ) return false;
    sod.IP96    = (struct P96IFace *)sod.IExec->GetInterface( sod.P96Base, "main", 1, NULL );
    if( !sod.IP96 )    return false;
    
//    IDOS->FPrintf( IDOS->Output(), "[ref_soft.so] Got P96!\n" );
//    IDOS->FFlush( IDOS->Output() );
    
    sod.IntuitionBase = sod.IExec->OpenLibrary("intuition.library", 53L);
    if (!sod.IntuitionBase) return false;
    sod.IIntuition = (struct IntuitionIFace *)sod.IExec->GetInterface( sod.IntuitionBase, "main", 1, NULL );
    if( !sod.IIntuition ) return false;

    sod.GfxBase = sod.IExec->OpenLibrary("graphics.library", 53L);
    if (!sod.GfxBase) return false;
    sod.IGraphics = (struct GraphicsIFace *)sod.IExec->GetInterface( sod.GfxBase, "main", 1, NULL );
    if( !sod.IGraphics ) return false;

    return true;
}

void SWimp_CloseLibs(void)
{
    if (sod.IP96) sod.IExec->DropInterface( (struct Interface *)sod.IP96 );
    if (sod.P96Base) sod.IExec->CloseLibrary( sod.P96Base );
    if (sod.IIntuition) sod.IExec->DropInterface( (struct Interface *)sod.IIntuition );
    if (sod.IntuitionBase) sod.IExec->CloseLibrary( sod.IntuitionBase );
    if (sod.IGraphics) sod.IExec->DropInterface( (struct Interface *)sod.IGraphics );
    if (sod.GfxBase) sod.IExec->CloseLibrary( sod.GfxBase );

    sod.IP96 = NULL;
    sod.P96Base = NULL;
    sod.IIntuition = NULL;
    sod.IntuitionBase = NULL;
    sod.IGraphics = NULL;
    sod.GfxBase = NULL;
    libs_init = false;
}

/*
** VID_CreateWindow
*/
#define WINDOW_CLASS_NAME "Quake 2"


ULONG GetMode(int w, int h);

qboolean VID_CreateWindow( int width, int height, qboolean fullscreen)
{
	int fullscreen_override;
	cvar_t  *vid_xpos,
		*vid_ypos,
		*vid_fullscreen,
		*sw_pubscreen;
	ULONG   wflgs;
	Tag     screentag;
	Tag     titletag;
	int     x,y;
	
	fullscreen_override = false;

	MAYBE_OPEN_LIBS;

	vid_xpos = ri.Cvar_Get ("vid_xpos", "0", 0);
	vid_ypos = ri.Cvar_Get ("vid_ypos", "0", 0);
	vid_fullscreen = ri.Cvar_Get ("vid_fullscreen", "0", CVAR_ARCHIVE );
	sw_pubscreen = ri.Cvar_Get("sw_pubscreen", "", CVAR_ARCHIVE);

	if ( vid_fullscreen->value || fullscreen)
	{
	    ULONG ModeID;
	    ULONG errorCode;
		struct TagItem ModeTags[3];

	ModeTags[0].ti_Tag=BIDTAG_NominalWidth;
	ModeTags[0].ti_Data = width;
	ModeTags[1].ti_Tag=BIDTAG_NominalHeight;
	ModeTags[1].ti_Data = height;
	ModeTags[2].ti_Tag=BIDTAG_Depth;
	ModeTags[2].ti_Data = 8;
//	ModeID=sod.IGraphics->BestModeIDA(ModeTags);
	
	ModeID = GetMode(width, height);	
	//	ModeID = sod.IGraphics->BestModeIDA(ModeTags);

      /*  ModeID = sod.IP96->p96BestModeIDTags( P96BIDTAG_NominalWidth,  width,
                                              P96BIDTAG_NominalHeight, height,
                                              P96BIDTAG_Depth,         8,
                                              TAG_DONE );*/
		
		g_pScreen = 0;
		if (ModeID!=0)
		{
		g_pScreen = sod.IIntuition->OpenScreenTags(NULL,
		SA_DisplayID,           ModeID,
		SA_Width,               width,
		SA_Height,              height,
		SA_Depth,               8,
		SA_ShowTitle,           FALSE,
		SA_Quiet,               TRUE,
		SA_FullPalette,         TRUE,
		SA_ErrorCode,           (ULONG)&errorCode,
	    TAG_DONE);
		}

	    if (!g_pScreen)
	    {
		ri.Con_Printf(PRINT_ALL, "OpenScreen: ");
		switch(errorCode)
		{
		    case OSERR_NOMONITOR:
			ri.Con_Printf(PRINT_ALL, "Monitor for display mode not available\n");
			break;
		    case OSERR_NOCHIPS:
			ri.Con_Printf(PRINT_ALL, "Custom chips for display mode missing\n");
			break;
		    case OSERR_NOMEM:
			ri.Con_Printf(PRINT_ALL, "Out of system memory\n");
			break;
		    case OSERR_NOCHIPMEM:
			ri.Con_Printf(PRINT_ALL, "Out of chip memory\n");
			break;
		    case OSERR_PUBNOTUNIQUE:
			ri.Con_Printf(PRINT_ALL, "Public screen name already in use\n");
			break;
		    case OSERR_UNKNOWNMODE:
			ri.Con_Printf(PRINT_ALL, "Didn't recognize display mode requested\n");
			break;
		    case OSERR_TOODEEP:
			ri.Con_Printf(PRINT_ALL, "This hardware cannot display screens of that depth\n");
			break;
		    case OSERR_ATTACHFAIL:
			ri.Con_Printf(PRINT_ALL, "An illegal attachment of screens was requested\n");
			break;
		    default:
			ri.Con_Printf(PRINT_ALL, "Error code %s\n", errorCode);
			break;
		}
		if (g_pScreen == 0)
		{
					g_pScreen = sod.IIntuition->LockPubScreen(NULL);

					if (g_pScreen == NULL)
					{
						g_pScreen = sod.IIntuition->LockPubScreen(NULL); // Fall back to default
						if (g_pScreen == NULL) return false;
					}
					screen_public = true;
					fullscreen_override = true;
		}
		if (g_pScreen == 0) return false;
	   
	    if (!fullscreen_override) screen_public = false;
		}
	}
	else
	{
	    if (sw_pubscreen->string && strlen(sw_pubscreen->string))
		g_pScreen = sod.IIntuition->LockPubScreen((UBYTE *)sw_pubscreen->string);
	    else
		g_pScreen = sod.IIntuition->LockPubScreen(NULL);

	    if (g_pScreen == NULL)
	    {
		g_pScreen = sod.IIntuition->LockPubScreen(NULL); // Fall back to default
		if (g_pScreen == NULL) return false;
	    }
	    screen_public = true;
	}

	if (vid_xpos && vid_xpos->value)
	    x = vid_xpos->value;
	else
	    x = 0;

	if (vid_ypos && vid_ypos->value)
	    y = vid_ypos->value;
	else
	    y = 0;


	// At this point g_pScreen is either a pubscreen pointer or a custom screen
	if (screen_public == true)
	{
	    screentag = WA_PubScreen;
	    titletag = WA_Title;
	    wflgs = WFLG_REPORTMOUSE
		  | WFLG_NOCAREREFRESH
		  | WFLG_ACTIVATE
		  | WFLG_RMBTRAP
		  | WFLG_DRAGBAR
		  | WFLG_DEPTHGADGET
		  | WFLG_CLOSEGADGET
		  ;
	}
	else
	{
	    screentag = WA_CustomScreen;
	    titletag = TAG_IGNORE;
	    wflgs = WFLG_BORDERLESS
		  | WFLG_REPORTMOUSE
		  | WFLG_NOCAREREFRESH
		  | WFLG_ACTIVATE
		  | WFLG_RMBTRAP
		  ;
		  x = 0;
		  y = 0;
	}

	g_pWindow = sod.IIntuition->OpenWindowTags(NULL,
	    WA_Left,            (ULONG)x,
	    WA_Top,             (ULONG)y,
	    WA_Flags,           wflgs,
	    WA_InnerWidth,      width,
	    WA_InnerHeight,     height,
	    screentag,          (ULONG)g_pScreen,
	    titletag,           (ULONG)"Quake 2",
	    WA_ScreenTitle,     (ULONG)"Quake 2 - Help / Scroll Lock captures mouse",
	TAG_DONE);
	


	if (!g_pWindow) ri.Sys_Error (ERR_FATAL, "Couldn't create window");

	g_pFriend = sod.IGraphics->AllocBitMap( width, height, g_pWindow->RPort->BitMap->Depth, BMF_DISPLAYABLE, g_pWindow->RPort->BitMap );
	if (!g_pFriend) ri.Sys_Error (ERR_FATAL, "Couldn't create bitmap");



	// let the sound and input subsystems know about the new window
	ri.Vid_NewWindow (width, height);
	SWimp_SetPalette(first_palette);

	return true;
}

/*
** SWimp_Init
**
** This routine is responsible for initializing the implementation
** specific stuff in a software rendering subsystem.
*/
int SWimp_Init( void *hInstance, void *wndProc )
{
    sod.IExec         = (struct ExecIFace *)(*(struct ExecBase **)4)->MainInterface;
    sod.P96Base       = NULL;
    sod.IP96          = NULL;
    sod.GfxBase       = NULL;
    sod.IGraphics     = NULL;
    sod.IntuitionBase = NULL;
    sod.IIntuition    = NULL;
    MAYBE_OPEN_LIBS;
    return true;
}

/*
** SWimp_InitGraphics
**
** This initializes the software refresh's implementation specific
** graphics subsystem.  In the case of Windows it creates DIB or
** DDRAW surfaces.
**
** The necessary width and height parameters are grabbed from
** vid.width and vid.height.
*/
static qboolean SWimp_InitGraphics( qboolean fullscreen )
{
	// free resources in use
	SWimp_Shutdown ();

	// create a new window
	VID_CreateWindow (vid.width, vid.height, fullscreen);

	vid.buffer      = malloc(vid.width * vid.height);
	vid.rowbytes    = vid.width;
	return true;
}

/*
** SWimp_EndFrame
**
** This does an implementation specific copy from the backbuffer to the
** front buffer.  In the Amiga case it uses LockBitMap to get the bitmap pointer,
** and then tries to either update directly in case of LUT8 or indirectly via a
** lookup table otherwise.
*/
void SWimp_EndFrame(void)
{
    struct BitMap *bm;
    if( !g_pWindow || !g_pFriend ) return;
    bm = g_pFriend;
    
    if( sod.IP96->p96GetBitMapAttr( bm, P96BMA_ISP96 ) )
    {
      int i, j;
      struct RenderInfo rinf;
      LONG handle;
      
      switch( sod.IP96->p96GetBitMapAttr( bm, P96BMA_BITSPERPIXEL ) )
      {
        case 8:
          handle = sod.IP96->p96LockBitMap( bm, (UBYTE *)&rinf, sizeof( struct RenderInfo ) );
          if( handle )
          {
            uint8 *dptr;
            uint8 *sptr;
            dptr = (uint8 *)rinf.Memory;
            sptr = vid.buffer;
            for( i=0; i<vid.height; i++ )
            {
              sod.IExec->CopyMem( sptr, dptr, vid.width );
              dptr += rinf.BytesPerRow;
              sptr += vid.rowbytes;
            }
            sod.IP96->p96UnlockBitMap( bm, handle );
          }
          break;
        
        case 15:
        case 16:
          handle = sod.IP96->p96LockBitMap( bm, (UBYTE *)&rinf, sizeof( struct RenderInfo ) );
          if( handle )
          {
            uint16 *dptr;
            uint8  *sptr;
            dptr = (uint16 *)rinf.Memory;
            sptr = vid.buffer;
            for( i=0; i<vid.height; i++ )
            {
              for( j=0; j<vid.width; j++ )
                *(dptr++) = word_pal[*(sptr++)];
              dptr += (rinf.BytesPerRow/2)-vid.width;
              sptr += vid.rowbytes - vid.width;
            }
            sod.IP96->p96UnlockBitMap( bm, handle );
          }
          break;

        case 24:
          handle = sod.IP96->p96LockBitMap( bm, (UBYTE *)&rinf, sizeof( struct RenderInfo ) );
          if( handle )
          {
            uint8 *dptr;
            uint8 *sptr;
            dptr = (uint8 *)rinf.Memory;
            sptr = vid.buffer;
            for( i=0; i<vid.height; i++ )
            {
              for( j=0; j<vid.width; j++ )
              {
                *(dptr++) = long_pal[*sptr]>>16;
                *(dptr++) = long_pal[*sptr]>>8;
                *(dptr++) = long_pal[*sptr];
                sptr++;
              }
              dptr += rinf.BytesPerRow-vid.width*3;
              sptr += vid.rowbytes - vid.width;
            }
            sod.IP96->p96UnlockBitMap( bm, handle );
          }
          break;
          
        case 32:
          handle = sod.IP96->p96LockBitMap( bm, (UBYTE *)&rinf, sizeof( struct RenderInfo ) );
          if( handle )
          {
            uint32 *dptr;
            uint8  *sptr;
            dptr = (uint32 *)rinf.Memory;
            sptr = vid.buffer;
            for( i=0; i<vid.height; i++ )
            {
              for( j=0; j<vid.width; j++ )
                *(dptr++) = long_pal[*(sptr++)];
              dptr += (rinf.BytesPerRow/4)-vid.width;
              sptr += vid.rowbytes - vid.width;
            }
            sod.IP96->p96UnlockBitMap( bm, handle );
          }
          break;

      }
      
      sod.IGraphics->BltBitMapRastPort( bm, 0, 0, g_pWindow->RPort, g_pWindow->BorderLeft, g_pWindow->BorderTop, vid.width, vid.height, 0x0c0 );
	}
/*
	else
	{
	    int i,j;
	    UBYTE *buffer = vid.buffer;
	    for (i = 0; i < vid.height; i++)
	    {
		UBYTE *b1 = buffer;
		UBYTE *b2 = base;
		for (j=0; j < vid.width; j++)
		{
		    *(UWORD *)base = word_pal[*buffer++];
		    base += BytesPerPix;
		}
		buffer = b1 + vid.rowbytes;
		base   = b2 + BytesPerRow;
	    }
	}
	UnLockBitMap(handle);
    }
    else
    {
	ri.Con_Printf(PRINT_ALL, "SWimp_EndFrame: No CyberGraphX-Compatible bitmap\n");
    }
*/
}

/*
** SWimp_SetMode
*/
rserr_t SWimp_SetMode( int *pwidth, int *pheight, int mode, qboolean fullscreen )
{
	const char *win_fs[] = { "W", "FS" };
	rserr_t retval = rserr_ok;

	ri.Con_Printf (PRINT_ALL, "setting mode %d:", mode );

	if ( !ri.Vid_GetModeInfo( pwidth, pheight, mode ) )
	{
		ri.Con_Printf( PRINT_ALL, " invalid mode\n" );
		return rserr_invalid_mode;
	}

	ri.Con_Printf( PRINT_ALL, " %d %d %s\n", *pwidth, *pheight, win_fs[fullscreen] );

	if ( fullscreen )
	{
		if ( !SWimp_InitGraphics( 1 ) )
		{
			if ( SWimp_InitGraphics( 0 ) )
			{
				// mode is legal but not as fullscreen
				fullscreen = 0;
				retval = rserr_invalid_fullscreen;
			}
			else
			{
				// failed to set a valid mode in windowed mode
				retval = rserr_unknown;
			}
		}
	}
	else
	{
		// failure to set a valid mode in windowed mode
		if ( !SWimp_InitGraphics( fullscreen ) )
		{
			return rserr_unknown;
		}
	}

	sw_state.fullscreen = fullscreen;
	R_GammaCorrectAndSetPalette( ( const unsigned char * ) d_8to24table );

	return retval;
}

void FS_SetPalette(const unsigned char *palette)
{
    int i,j;
    ULONG xpalette[300*3];
    xpalette[0] = 256<<16;
    if (!g_pScreen) return;

    j=1;

    for (i=0; i<256; i++)
    {
	char r,g,b;
	r = palette[i*4];
	g = palette[i*4+1];
	b = palette[i*4+2];
	xpalette[j++] = (r << 24) | 0xFFFFFF;
	xpalette[j++] = (g << 24) | 0xFFFFFF;
	xpalette[j++] = (b << 24) | 0xFFFFFF;
    }

    xpalette[j] = 0;

    sod.IGraphics->LoadRGB32(&(g_pScreen->ViewPort), xpalette);
}


/*
** SWimp_SetPalette
**
** System specific palette setting routine.  A NULL palette means
** to use the existing palette.  The palette is expected to be in
** a padded 4-byte xRGB format.
*/
void SWimp_SetPalette( const unsigned char *palette )
{
    ULONG PixFmt;
    uint32 bpp, fmt;
    struct BitMap *bm;
    int i;
    
    if (!g_pWindow)
    {
	memcpy(first_palette, palette, 4*256);
	return;
    }

    bm = g_pWindow->RPort->BitMap;
    
    if ( !palette )
	palette = ( const unsigned char * ) sw_state.currentpalette;
	
	bpp = sod.IP96->p96GetBitMapAttr( bm, P96BMA_BITSPERPIXEL );
	fmt = sod.IP96->p96GetBitMapAttr( bm, P96BMA_RGBFORMAT );

    switch( bpp )
    {
      case 8:
        FS_SetPalette( palette );
        break;
      
      case 15:
      case 16:
        for( i=0; i<256; i++ )
          word_pal[i] = sod.IP96->p96EncodeColor( fmt, (palette[i*4]<<16)|(palette[i*4+1]<<8)|palette[i*4+2] );
        break;

      case 24:
      case 32:
        for( i=0; i<256; i++ )
          long_pal[i] = sod.IP96->p96EncodeColor( fmt, (palette[i*4]<<16)|(palette[i*4+1]<<8)|palette[i*4+2] );
        break;

      default:
	    ri.Con_Printf(PRINT_ALL, "Unsupported pixmap format\n");
	    break;
	}
}

/*
** SWimp_Shutdown
**
** System specific graphics subsystem shutdown routine.  Destroys
** DIBs or DDRAW surfaces as appropriate.
*/
void SWimp_Shutdown( void )
{
    ri.Con_Printf( PRINT_ALL, "Shutting down SW imp\n" );

    if( g_pFriend ) sod.IGraphics->FreeBitMap( g_pFriend );
    g_pFriend = 0;

    if (g_pWindow) sod.IIntuition->CloseWindow(g_pWindow);
    g_pWindow = 0;

    if (g_pScreen)
	if (screen_public == false) sod.IIntuition->CloseScreen(g_pScreen);
	else sod.IIntuition->UnlockPubScreen(NULL, g_pScreen);
    g_pScreen = 0;

    screen_public = false;
    
}

/*
** SWimp_AppActivate
*/
void SWimp_AppActivate( qboolean active )
{
}

//===============================================================================

struct Window *GetWindowHandle(void)
{
    return g_pWindow;
}

extern refexport_t GetRefAPI(refimport_t rimp);

ULONG GetMode(int w, int h)
{
    ULONG modeID, error, result;
    struct DisplayInfo dpinfo;
    struct NameInfo nminfo;
    struct DimensionInfo diminfo;
    DisplayInfoHandle displayhandle;

    ULONG bestMode = INVALID_ID, bestW=0xffffffff, bestH = 0xffffffff;
    ULONG nW, nH;


    modeID = INVALID_ID;
    
    sod.IGraphics->GetDisplayInfoData(NULL, (UBYTE *)&dpinfo, sizeof(struct DisplayInfo),
	    DTAG_DISP, LORES_KEY);

    while ((modeID = sod.IGraphics->NextDisplayInfo(modeID)) != INVALID_ID)
    {
	if (sod.IP96->p96GetModeIDAttr( modeID, P96IDA_ISP96 ) == FALSE )
	    continue;

	error = sod.IGraphics->ModeNotAvailable(modeID);
	    
	if (error == 0)
	{
	    displayhandle = sod.IGraphics->FindDisplayInfo(modeID);
	    if (displayhandle)
	    {
		result = sod.IGraphics->GetDisplayInfoData(displayhandle, (UBYTE *)&diminfo,
			    sizeof(struct DimensionInfo), DTAG_DIMS, 0);
		if (result)
		{
		    if (diminfo.MaxDepth == 8)
		    {
			nW = diminfo.Nominal.MaxX - diminfo.Nominal.MinX + 1;
			nH = diminfo.Nominal.MaxY - diminfo.Nominal.MinY + 1;

			if (nW == h && nH == w)
			{
			    // Exact match
			    return modeID;
			}
			else
			{
			    // Check if the widht and height are big enough to even take the screen
			    if (nW >= w && nH >= h)
			    {
				// Yes. Check if they match the requested size better
				if (nW <= bestW && nH <= bestH)
				{
				    // They do
				    bestW = nW;
				    bestH = nH;
				    bestMode = modeID;
				}
			    }
			}
		    } // if (diminfo...
		} // if (result...
	    } // if (displayhandle...
	} // if (error...
    } // while


    // Here, we either have a mode that roughly matches, or none
    return bestMode;
}

