#include <stdio.h>
#include "../client/ref.h"
#ifdef STORM
#include "keys.h"
#else
#include "client/keys.h"
#endif

#include <exec/exec.h>
#include <proto/exec.h>
#include <proto/keymap.h>
#include <intuition/intuition.h>
#include <proto/intuition.h>
#include <graphics/gfx.h>
#include <proto/graphics.h>
#include <proto/elf.h>
#include <exec/tasks.h>
#include <sys/stat.h>
#include <dlfcn.h>
//#include "dll.h"


cvar_t *nostdout;
cvar_t *serialout = 0;
extern cvar_t          *vid_fullscreen;

unsigned int sys_frame_time;
unsigned int sys_msg_time;

extern refexport_t     re;
extern struct Window * (*GetWindowHandle)(void);

struct Library *Q2_ElfBase = NULL;
struct ElfIFace *Q2_IElf = NULL;

uid_t saved_euid;
qboolean stdin_active = true;

char __attribute__((used)) stackcookie[] = "$STACK: 1500000";


static void *g_pCursor = 0;


// =======================================================================
// General routines
// =======================================================================

void Sys_ConsoleOutput (char *string)
{
	if (serialout && serialout->value)
	    IExec->DebugPrintF("%s", string);

	if (nostdout == 0)
		return;
	if (nostdout && nostdout->value)
		return;

	fputs(string, stdout);
}

void Sys_Printf (char *fmt, ...)
{
	va_list         argptr;
	char            text[1024];
	unsigned char           *p;

	va_start (argptr,fmt);
	vsprintf (text,fmt,argptr);
	va_end (argptr);

	if (strlen(text) > sizeof(text))
		Sys_Error("memory overwrite in Sys_Printf");

	if (nostdout == 0)
		return;

	if (nostdout && nostdout->value)
	    return;

	for (p = (unsigned char *)text; *p; p++) {
		*p &= 0x7f;
		if ((*p > 128 || *p < 32) && *p != 10 && *p != 13 && *p != 9)
			printf("[%02x]", *p);
		else
			putc(*p, stdout);
	}
}


// Q2_ prefixes avoid conflicts with shared objects that might
// also want to use these libraries.
struct Library *Q2_IntuitionBase = NULL;
struct Library *Q2_GfxBase = NULL;
struct Library *SocketBase = NULL;
struct Library *Q2_KeymapBase = NULL;

struct IntuitionIFace *Q2_IIntuition = NULL;
struct GraphicsIFace  *Q2_IGraphics = NULL;
struct SocketIFace    *ISocket = NULL;
struct KeymapIFace    *Q2_IKeymap = NULL;

struct Library *ElfBase = NULL;
struct ElfIFace *IElf = NULL;

void Sys_Leave(void)
{
    if (Q2_IIntuition)         IExec->DropInterface((struct Interface *)Q2_IIntuition);
    if (Q2_IntuitionBase)      IExec->CloseLibrary(Q2_IntuitionBase);
    if (Q2_IGraphics)          IExec->DropInterface((struct Interface *)Q2_IGraphics);
    if (Q2_GfxBase)            IExec->CloseLibrary(Q2_GfxBase);
    if (ISocket)               IExec->DropInterface((struct Interface *)ISocket);
    if (SocketBase)            IExec->CloseLibrary(SocketBase);
    if (Q2_IKeymap)            IExec->DropInterface((struct Interface *)Q2_IKeymap);
    if (Q2_KeymapBase)         IExec->CloseLibrary(Q2_KeymapBase);
    if (Q2_IElf)               IExec->DropInterface((struct Interface *)Q2_IElf);
    if (Q2_ElfBase)            IExec->CloseLibrary(Q2_ElfBase);

    Q2_IIntuition      = NULL;
    Q2_IntuitionBase   = NULL;
    Q2_IGraphics       = NULL;
    Q2_GfxBase         = NULL;
    ISocket            = NULL;
    SocketBase         = NULL;
    Q2_IKeymap         = NULL;
    Q2_KeymapBase      = NULL;
    Q2_IElf            = NULL;
    Q2_ElfBase         = NULL;
}

void Z_FreeAll();

void Sys_Quit (void)
{
  if (g_pCursor) IExec->FreeVec(g_pCursor);
  g_pCursor = 0;

	Sys_UnloadGame();
	VID_Shutdown();
	SNDDMA_Shutdown();
	IN_ShutdownJoystick();
	CL_Shutdown ();
	Qcommon_Shutdown ();
	Sys_Leave();
	Z_FreeAll();
	exit(0);
}

void Sys_Init(void)
{
	Q2_ElfBase       = IExec->OpenLibrary("elf.library",53L);
	ElfBase = Q2_ElfBase;
	Q2_IElf          = (struct ElfIFace *)IExec->GetInterface( Q2_ElfBase, "main", 1, NULL );
	IElf = Q2_IElf;
    Q2_IntuitionBase = IExec->OpenLibrary("intuition.library", 0L);
    Q2_IIntuition    = (struct IntuitionIFace *)IExec->GetInterface( Q2_IntuitionBase, "main", 1, NULL );
    Q2_GfxBase       = IExec->OpenLibrary("graphics.library", 0L);
    Q2_IGraphics     = (struct GraphicsIFace *)IExec->GetInterface( Q2_GfxBase, "main", 1, NULL );
    SocketBase       = IExec->OpenLibrary("bsdsocket.library", 0L);
    ISocket          = (struct SocketIFace *)IExec->GetInterface( SocketBase, "main", 1, NULL );
    Q2_KeymapBase    = IExec->OpenLibrary("keymap.library", 0);
    Q2_IKeymap       = (struct KeymapIFace *)IExec->GetInterface( Q2_KeymapBase, "main", 1, NULL );
}

void Sys_Error (char *error, ...)
{ 
    va_list     argptr;
    char        string[1024];

	CL_Shutdown ();
	Qcommon_Shutdown ();
	Sys_UnloadGame();
	VID_Shutdown();
	SNDDMA_Shutdown();
	IN_ShutdownJoystick();
    
    va_start (argptr,error);
    vsprintf (string,error,argptr);
    va_end (argptr);
	fprintf(stderr, "Error: %s\n", string);

	Sys_Leave();
	Z_FreeAll();
	exit(1);

} 

void Sys_Warn (char *warning, ...)
{ 
    va_list     argptr;
    char        string[1024];
    
    va_start (argptr,warning);
    vsprintf (string,warning,argptr);
    va_end (argptr);
	fprintf(stderr, "Warning: %s", string);
} 

/*
============
Sys_FileTime

returns -1 if not present
============
*/
int     Sys_FileTime (char *path)
{
	struct  stat    buf;
	
	if (stat (path,&buf) == -1)
		return -1;
	
	return buf.st_mtime;
}

char *Sys_ConsoleInput(void)
{
	return NULL;
}

/*****************************************************************************/

static void *game_library;

/*
=================
Sys_UnloadGame
=================
*/
void Sys_UnloadGame (void)
{
	if (game_library)
	  dlclose( game_library );
	game_library = NULL;
}

/*
=================
Sys_GetGameAPI

Loads the game dll
=================
*/
void *Sys_GetGameAPI (void *parms)
{
	void    *(*GetGameAPI) (void *);
	void    (*SetExeName) (char *name);

	char    name[MAX_OSPATH];
	char    curpath[MAX_OSPATH];
	char    *path;
	const char *gamename = "game.so";
	void    **SegList;

	if (game_library)
		Com_Error (ERR_FATAL, "Sys_GetGameAPI without Sys_UnloadingGame");

	getcwd(curpath, sizeof(curpath));

	Com_Printf("------- Loading %s -------", gamename);

	// now run through the search paths
	path = NULL;
	while (1)
	{
		path = FS_NextPath (path);
		if (!path)
			return NULL;            // couldn't find one anywhere
		sprintf (name, "%s/%s/%s", curpath, path, gamename);
		game_library = dlopen( name, RTLD_LAZY );
		if (game_library)
		{
			Com_Printf ("\nLoadLibrary (%s)\n",name);
			break;
		}
		else
		{
			Com_Printf("\nLoadLibrary (%s) failed\n", name);
			break;
		}
	}

    SetExeName = (void *)dlsym( game_library, "SetExeName" );
	if (SetExeName) SetExeName(COM_Argv(0));

    GetGameAPI = (void *)dlsym( game_library, "GetGameAPI" );
	if (!GetGameAPI)
	{
		Sys_UnloadGame ();              
		return NULL;
	}

	return GetGameAPI (parms);
}

/*****************************************************************************/

void Sys_AppActivate (void)
{
}

#define ENCODEKEY(code,qual) (int)(((int)code & 0x7F) | ((int)qual << 8))

static BOOL cursor_is_on = TRUE;
int recheckgrab=1;
static void HideCursor(void)
{
    struct Window *w;

    if (!g_pCursor)
        g_pCursor = IExec->AllocVecTags(24, AVT_ClearWithValue, 0, TAG_DONE);

    if (GetWindowHandle)
    {
        w = GetWindowHandle();
        if (w) Q2_IIntuition->SetPointer( w, g_pCursor, 1, 16, 0, 0);
    }
    
    cursor_is_on = FALSE;
}

static void ShowCursor(void)
{
    struct Window *w;

    if (GetWindowHandle)
    {
        w = GetWindowHandle();
        if (w) Q2_IIntuition->ClearPointer( w );
    }
    
    cursor_is_on = TRUE;
}


static BOOL mousegrabbed = FALSE;
void setmousegrab( struct Window *w, BOOL active )
{
	struct IBox grabBox = { w->BorderLeft,
	                        w->BorderTop,
	                        w->Width - w->BorderLeft - w->BorderRight,
	                        w->Height - w->BorderTop - w->BorderBottom };

//	printf( "w = %p, bt = %d\n", w, w->BorderTop );

	recheckgrab = 0;

	// Fullscreen?
	if( w->BorderTop == 0 )
	{
//		printf( "Fullscreen forcer\n" );
		HideCursor();
		mousegrabbed = FALSE;
		Q2_IIntuition->SetWindowAttrs( w, WA_GrabFocus, 0, sizeof( ULONG ) );
		return;
	}

//	printf( "Windowed set to %d\n", active );
	Q2_IIntuition->SetWindowAttrs( w, WA_MouseLimits, &grabBox, sizeof( grabBox ) );
	Q2_IIntuition->SetWindowAttrs( w, WA_GrabFocus, active ? 100 : 0, sizeof( ULONG ) );

	if( active )
	{
		HideCursor();
	} else {
		ShowCursor();
	}

	mousegrabbed = active;
}

unsigned int mouse_wheel_used = 0; //18-AUG-02
	int aborted_cinematic_hack=0;
	extern int special_cinema;

void Sys_SendKeyEvents (void)
{
    struct IntuiMessage *imsg;
    struct Window *w = NULL;
	  extern struct MsgPort *g_pMessagePort;
    extern unsigned int inittime;

    if (GetWindowHandle)
    {
	w = GetWindowHandle();
	if (w) g_pMessagePort = w->UserPort;
	if (g_pMessagePort == 0)
	{
	    Q2_IIntuition->ModifyIDCMP(w, IDCMP_RAWKEY|IDCMP_MOUSEMOVE|IDCMP_MOUSEBUTTONS|IDCMP_DELTAMOVE|IDCMP_CLOSEWINDOW);
	    Q2_IIntuition->ReportMouse(TRUE, w);
	    g_pMessagePort = w->UserPort;
	}
    }

	// Regrab the mouse in windowed mode
	if( mousegrabbed && w && (w->BorderTop!=0) )
		Q2_IIntuition->SetWindowAttrs( w, WA_GrabFocus, 100, sizeof( ULONG ) );

    while (imsg = (struct IntuiMessage *)IExec->GetMsg(g_pMessagePort))
    {
	sys_msg_time=(imsg->Seconds-inittime)*1000+imsg->Micros/1000;
	switch(imsg->Class)
	{
	case IDCMP_RAWKEY:
		
						if (special_cinema) 
						{
							special_cinema = 0;
							aborted_cinematic_hack = 1;
						}
	    if ((0x7F & imsg->Code) == 122 || (0x7F & imsg->Code) == 123)
	    {
		mouse_wheel_used = 1; //18-AUG-02

		switch(imsg->Code)
		{
		case 0x7A:
		    Key_Event(K_MWHEELUP, true, sys_msg_time, -1); break;
		case 0xFA:
		    Key_Event(K_MWHEELUP, false, sys_msg_time, -1); break;
		case 0x7B:
		    Key_Event(K_MWHEELDOWN, true, sys_msg_time, -1); break;
		case 0xFB:
		    Key_Event(K_MWHEELDOWN, false, sys_msg_time, -1); break;
		}
	    }

      if( ((imsg->Code)&0x7f) == 0x5f )
		  {
				if( w )
				{
  	      if ((imsg->Code)&0x80)
            setmousegrab( w, !mousegrabbed );
			  }
			  break;
			}

	    if ((imsg->Code)&0x80)
		Key_Event(MapKey((imsg->Code)&0x7f),false,sys_msg_time, ENCODEKEY(imsg->Code, imsg->Qualifier));
	    else
		Key_Event(MapKey(imsg->Code),true,sys_msg_time, ENCODEKEY(imsg->Code, imsg->Qualifier));
	    break;
	case IDCMP_MOUSEBUTTONS:
	    switch(imsg->Code)
	    {
	    case IECODE_LBUTTON:
						if (special_cinema) 
						{
							special_cinema = 0;
							aborted_cinematic_hack = 1;
						}
		Key_Event(K_MOUSE1,true,sys_msg_time, -1);
		break;
	    case (IECODE_LBUTTON | IECODE_UP_PREFIX):
		Key_Event(K_MOUSE1,false,sys_msg_time, -1);
		break;
	    case IECODE_RBUTTON:
		Key_Event(K_MOUSE1+1,true,sys_msg_time, -1);
		break;
	    case (IECODE_RBUTTON | IECODE_UP_PREFIX):
		Key_Event(K_MOUSE1+1,false,sys_msg_time, -1);
		break;
	    case IECODE_MBUTTON:
		Key_Event(K_MOUSE1+2,true,sys_msg_time, -1);
		break;
	    case (IECODE_MBUTTON | IECODE_UP_PREFIX):
		Key_Event(K_MOUSE1+2,false,sys_msg_time, -1);
		break;
	    }
	    break;
	case IDCMP_DELTAMOVE:
	case IDCMP_MOUSEMOVE:
	    IN_GetMouseMove(imsg);
	    break;
	case IDCMP_CLOSEWINDOW:
	    Cbuf_AddText("quit");
						if (special_cinema) 
						{
							special_cinema = 0;
							aborted_cinematic_hack = 1;
						}
	    break;
	}

	IExec->ReplyMsg((struct Message *)imsg);
    }

//18-AUG-02: workaround for bug in Topolino3 mouse driver
    if( (!imsg) && (mouse_wheel_used) )
    {
	Key_Event(K_MWHEELUP, false, sys_msg_time, -1);
	Key_Event(K_MWHEELDOWN, false, sys_msg_time, -1);
	mouse_wheel_used = 0;
    }

    // grab frame time
    sys_frame_time = Sys_Milliseconds();
}

int Sys_SendKeyEvents2 (void)
{
    struct IntuiMessage *imsg;
    struct Window *w = NULL;
	  extern struct MsgPort *g_pMessagePort;
    extern unsigned int inittime;

    if (GetWindowHandle)
    {
	w = GetWindowHandle();
	if (w) g_pMessagePort = w->UserPort;
	if (g_pMessagePort == 0)
	{
	    Q2_IIntuition->ModifyIDCMP(w, IDCMP_RAWKEY|IDCMP_MOUSEMOVE|IDCMP_MOUSEBUTTONS|IDCMP_DELTAMOVE|IDCMP_CLOSEWINDOW);
	    Q2_IIntuition->ReportMouse(TRUE, w);
	    g_pMessagePort = w->UserPort;
	}
    }

	// Regrab the mouse in windowed mode
	if( mousegrabbed && w && (w->BorderTop!=0) )
		Q2_IIntuition->SetWindowAttrs( w, WA_GrabFocus, 100, sizeof( ULONG ) );

    while (imsg = (struct IntuiMessage *)IExec->GetMsg(g_pMessagePort))
    {
	sys_msg_time=(imsg->Seconds-inittime)*1000+imsg->Micros/1000;
	switch(imsg->Class)
	{
	case IDCMP_RAWKEY:
		IExec->ReplyMsg((struct Message *)imsg);
		return 1;
	    break;
	case IDCMP_MOUSEBUTTONS:
		IExec->ReplyMsg((struct Message *)imsg);
	    return 1;
	    break;
	case IDCMP_DELTAMOVE:
	case IDCMP_MOUSEMOVE:
	    IN_GetMouseMove(imsg);
	    break;
	case IDCMP_CLOSEWINDOW:
	    Cbuf_AddText("quit");
		IExec->ReplyMsg((struct Message *)imsg);
		return 1;
	    break;
	}

	IExec->ReplyMsg((struct Message *)imsg);
    }

    // grab frame time
    sys_frame_time = Sys_Milliseconds();
}

/*****************************************************************************/

char *Sys_GetClipboardData(void)
{
	return NULL;
}


//#define MAGAZINTEST

int main (int argc, char **argv)
{
    int     time, oldtime, newtime;

#ifdef MAGAZINTEST
		char id2[100];
    int i;
    int invalid=0;

    CopyMem(0xF00010, id2, 8);
    for (i=0;i<8;i++)
    {
      if (i!=1)
      {
        if (id2[i]!=255) invalid=1;
      }
    }
    if (id2[1]!=65) invalid=1;
    if (invalid) exit(0);
#endif


    Qcommon_Init(argc, argv);

    nostdout = Cvar_Get("nostdout", "1", CVAR_ARCHIVE);
    serialout = Cvar_Get("serialout", "0", 0);

    Com_Printf("\n");
    Com_Printf("  \35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n");
    Com_Printf("\1         Quake \21\20 Amiga\n");
    Com_Printf("  \35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n");
    Com_Printf("           Ported by:\n");
    Com_Printf("         Steffen Haeuser etc.\n");
    Com_Printf("  \35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n");
    Com_Printf("\n");

    oldtime = Sys_Milliseconds ();
    while (1)
    {
	do
	{
	    newtime = Sys_Milliseconds ();
	    time = newtime - oldtime;
	} while (time < 1);
	Qcommon_Frame (time);
	oldtime = newtime;
    }

}

void Sys_CopyProtect(void)
{
}

void isort(void *pArray, unsigned int nElements, unsigned int nSize, int (*compare)(const void *a, const void *b))
{
    int i, j;
    unsigned char *pCharArray = (unsigned char *)pArray;
    unsigned char *pRef = malloc(nSize);

    for (i = 1; i < nElements; i++)
    {
	memcpy(pRef, pCharArray + nSize * i, nSize);
	j = i;

	while (j > 0 && compare(pCharArray + (j-1)*nSize, pRef) > 0)
	{
	    memcpy(pCharArray + j*nSize, pCharArray + (j-1)*nSize, nSize);
	    j--;
	}

	memcpy(pCharArray + j*nSize, pRef, nSize);
    }

    free(pRef);
}

void Sys_ConvertFilename(char *filename)
{
    while (*filename)
    {
	if (*filename == '\\')
	    *filename = '/';
	filename++;
    }
}

FILE * Sys_fopen(char *filename, char *access)
{
    char name[MAX_OSPATH];

    strcpy(name, filename);

    Sys_ConvertFilename(name);

    return fopen(name, access);
}



int Sys_MapRawKey(int nRawkey)
{
    struct InputEvent ie;
    WORD nResult;
    char cBuffer[100];
    UWORD nCode = nRawkey & 0xFF;
    UWORD nQualifier = (nRawkey >> 8) % 0xFFFF;

    switch (nCode)
    {
	case 65:
	    return -1;

	default:
	    ie.ie_Class         = IECLASS_RAWKEY;
	    ie.ie_SubClass      = 0;
	    ie.ie_Code          = nCode;
	    ie.ie_Qualifier     = nQualifier;
	    ie.ie_EventAddress  = NULL;

	    nResult = Q2_IKeymap->MapRawKey(&ie, cBuffer, 100, 0);

	    if (nResult != 1)
		return -1;
	    else
		return cBuffer[0];
    }

    return -1;
}
