/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "../client/client.h"
#include "../client/qmenu.h"

#define REF_SOFT        0
#define REF_OPENGL      1
//#define REF_3DFX        2
//#define REF_POWERVR     3
//#define REF_VERITE      4

extern cvar_t *vid_ref;
extern cvar_t *vid_fullscreen;
extern cvar_t *vid_gamma;
extern cvar_t *scr_viewsize;

static cvar_t *gl_mode;
static cvar_t *gl_driver;
static cvar_t *gl_picmip;
static cvar_t *gl_ext_palettedtexture;
static cvar_t *gl_finish;
static cvar_t *gl_lockmode;

static cvar_t *sw_mode;
static cvar_t *sw_stipplealpha;

extern void M_ForceMenuOff( void );

/*
====================================================================

MENU INTERACTION

====================================================================
*/
#define SOFTWARE_MENU 0
#define OPENGL_MENU   1
#define OPENGLNOLRU_MENU 2

static menuframework_s  s_software_menu;
static menuframework_s  s_opengl_menu;
static menuframework_s  s_openglnolru_menu;
static menuframework_s *s_current_menu;
static int                              s_current_menu_index;

static menulist_s               s_mode_list[3];
static menulist_s               s_ref_list[3];
static menuslider_s             s_tq_slider;
static menuslider_s             s_screensize_slider[3];
static menuslider_s             s_brightness_slider[3];
static menulist_s               s_fs_box[3];
static menulist_s               s_stipple_box;
static menulist_s               s_paletted_texture_box;
static menulist_s               s_finish_box;
static menulist_s               s_lockmode_box;
static menuaction_s             s_cancel_action[3];
static menuaction_s             s_defaults_action[3];

static void DriverCallback( void *unused )
{
	if ( s_ref_list[s_current_menu_index].curvalue == 0 )
	{
		s_current_menu = &s_software_menu;
		s_current_menu_index = 0;
    s_ref_list[0].curvalue = 0;
    s_ref_list[1].curvalue = 0;
    s_ref_list[2].curvalue = 1;
	}
	else  if (s_ref_list[s_current_menu_index].curvalue == 1)
	{
		s_current_menu = &s_opengl_menu;
		s_current_menu_index = 1;
    s_ref_list[0].curvalue = 0;
    s_ref_list[1].curvalue = 1;
    s_ref_list[2].curvalue = 2;
	}
  else
  {
    s_current_menu = &s_openglnolru_menu;
    s_current_menu_index = 2;
    s_ref_list[0].curvalue = 2;
    s_ref_list[1].curvalue = 1;
    s_ref_list[2].curvalue = 2;
  }
}

static void ScreenSizeCallback( void *s )
{
	menuslider_s *slider = ( menuslider_s * ) s;

	Cvar_SetValue( "viewsize", slider->curvalue * 10 );
}

static void BrightnessCallback( void *s )
{
	menuslider_s *slider = ( menuslider_s * ) s;

	if ( s_current_menu_index == SOFTWARE_MENU )
		s_brightness_slider[1].curvalue = s_brightness_slider[0].curvalue;
	else if (s_current_menu_index == OPENGL_MENU)
		s_brightness_slider[0].curvalue = s_brightness_slider[1].curvalue;
    else s_brightness_slider[2].curvalue = s_brightness_slider[2].curvalue;

	if ( stricmp( vid_ref->string, "soft" ) == 0 )
	{
		float gamma = ( 0.8 - ( slider->curvalue/10.0 - 0.5 ) ) + 0.5;

		Cvar_SetValue( "vid_gamma", gamma );
	}
}

static void ResetDefaults( void *unused )
{
	VID_MenuInit();
}

static void ApplyChanges( void *unused )
{
	float gamma;
	int i;

	/*
	** make values consistent
	*/
	//s_fs_box[!s_current_menu_index].curvalue = s_fs_box[s_current_menu_index].curvalue;
	s_brightness_slider[!s_current_menu_index].curvalue = s_brightness_slider[s_current_menu_index].curvalue;

	if ( s_ref_list[s_current_menu_index].curvalue == 0 )
	{
    s_ref_list[0].curvalue = 0;
    s_ref_list[1].curvalue = 0;
    s_ref_list[2].curvalue = 1;
    s_fs_box[1].curvalue = s_fs_box[s_current_menu_index].curvalue;
    s_fs_box[2].curvalue = s_fs_box[s_current_menu_index].curvalue;
  	s_brightness_slider[1].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
    s_brightness_slider[2].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
	}
	else if (s_ref_list[s_current_menu_index].curvalue == 1)
	{
    s_ref_list[0].curvalue = 0;
    s_ref_list[1].curvalue = 1;
    s_ref_list[2].curvalue = 2;
    s_fs_box[0].curvalue = s_fs_box[s_current_menu_index].curvalue;
    s_fs_box[2].curvalue = s_fs_box[s_current_menu_index].curvalue;
  	s_brightness_slider[0].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
    s_brightness_slider[2].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
	}
  else
  {
    s_ref_list[0].curvalue = 2;
    s_ref_list[1].curvalue = 1;
    s_ref_list[2].curvalue = 2;
    s_fs_box[0].curvalue = s_fs_box[s_current_menu_index].curvalue;
    s_fs_box[1].curvalue = s_fs_box[s_current_menu_index].curvalue;
    s_brightness_slider[0].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
    s_brightness_slider[1].curvalue = s_brightness_slider[s_current_menu_index].curvalue;
  }
  /*
	** invert sense so greater = brighter, and scale to a range of 0.5 to 1.3
	*/
	gamma = ( 0.8 - ( s_brightness_slider[s_current_menu_index].curvalue/10.0 - 0.5 ) ) + 0.5;

	Cvar_SetValue( "vid_gamma", gamma );
	Cvar_SetValue( "sw_stipplealpha", s_stipple_box.curvalue );
	Cvar_SetValue( "gl_picmip", 3 - s_tq_slider.curvalue );
	Cvar_SetValue( "vid_fullscreen", s_fs_box[s_current_menu_index].curvalue );
	Cvar_SetValue( "gl_ext_palettedtexture", s_paletted_texture_box.curvalue );
	Cvar_SetValue( "gl_finish", s_finish_box.curvalue );
	Cvar_SetValue( "sw_mode", s_mode_list[SOFTWARE_MENU].curvalue );
	if (s_ref_list[s_current_menu_index].curvalue==REF_OPENGL) Cvar_SetValue( "gl_mode", s_mode_list[OPENGL_MENU].curvalue );
  	else if (s_ref_list[s_current_menu_index].curvalue==2) Cvar_SetValue("gl_mode", s_mode_list[2].curvalue);

	switch ( s_ref_list[s_current_menu_index].curvalue )
	{
	case REF_SOFT:
		Cvar_Set( "vid_ref", "soft" );
		break;
	case REF_OPENGL:
		Cvar_Set( "vid_ref", "egl" );
		Cvar_Set( "gl_driver", "opengl32" );
		break;
  case 2:
    Cvar_Set( "vid_ref", "gl");
    Cvar_Set( "gl_driver", "opengl32");
    break;
	}

	Cvar_Set( "gl_lockmode", s_lockmode_box.curvalue ? "SMART" : "MANUAL" );

	/*
	** update appropriate stuff if we're running OpenGL and gamma
	** has been modified
	*/
	if (( stricmp( vid_ref->string, "gl" ) == 0 )
      ||(stricmp(vid_ref->string,"egl") == 0))
	{
		if ( vid_gamma->modified )
		{
			vid_ref->modified = true;
		}

		if ( gl_driver->modified )
			vid_ref->modified = true;
	}

	M_ForceMenuOff();
}

static void CancelChanges( void *unused )
{
	extern void M_PopMenu( void );

	M_PopMenu();
}

/*
** VID_MenuInit
*/
void VID_MenuInit( void )
{
	static const char *resolutions[] = 
	{
		"[320 240  ]",
		"[400 300  ]",
		"[512 384  ]",
		"[640 480  ]",
		"[800 600  ]",
		"[960 720  ]",
		"[1024 768 ]",
		"[1152 864 ]",
		"[1280 720 ]",
		"[1280 960 ]",
		"[1600 1200]",
		"[1920 1080]",
		0
	};
	static const char *refs[] =
	{
		"[software      ]",
		"[GL4ES         ]",
		"[MiniGL        ]",
		0
	};
	static const char *yesno_names[] =
	{
		"no",
		"yes",
		0
	};
	static const char *lockmode_names[] =
	{
		"manual",
		"smart",
		0
	};

	int i;

	if ( !gl_driver )
		gl_driver = Cvar_Get( "gl_driver", "opengl32", 0 );
	if ( !gl_picmip )
		gl_picmip = Cvar_Get( "gl_picmip", "0", 0 );
	if ( !gl_mode )
		gl_mode = Cvar_Get( "gl_mode", "3", 0 );
	if ( !sw_mode )
		sw_mode = Cvar_Get( "sw_mode", "0", 0 );
	if ( !gl_ext_palettedtexture )
		gl_ext_palettedtexture = Cvar_Get( "gl_ext_palettedtexture", "1", CVAR_ARCHIVE );
	if ( !gl_finish )
		gl_finish = Cvar_Get( "gl_finish", "0", CVAR_ARCHIVE );
	if ( !gl_lockmode )
		gl_lockmode = Cvar_Get( "gl_lockmode", "SMART", CVAR_ARCHIVE );

	if ( !sw_stipplealpha )
		sw_stipplealpha = Cvar_Get( "sw_stipplealpha", "0", CVAR_ARCHIVE );

	s_mode_list[SOFTWARE_MENU].curvalue = sw_mode->value;
	s_mode_list[OPENGL_MENU].curvalue = gl_mode->value;
  s_mode_list[2].curvalue = gl_mode->value;

	if ( !scr_viewsize )
		scr_viewsize = Cvar_Get ("viewsize", "100", CVAR_ARCHIVE);

	s_screensize_slider[SOFTWARE_MENU].curvalue = scr_viewsize->value/10;
	s_screensize_slider[OPENGL_MENU].curvalue = scr_viewsize->value/10;
    s_screensize_slider[2].curvalue = scr_viewsize->value/10;

	if ( strcmp( vid_ref->string, "soft" ) == 0 )
	{
		s_current_menu_index = SOFTWARE_MENU;
		s_ref_list[0].curvalue = s_ref_list[1].curvalue = REF_SOFT;
  	s_ref_list[1].curvalue = 1;
    s_ref_list[2].curvalue = 2;
	}
	else if ( strcmp( vid_ref->string, "egl" ) == 0 )
	{
		s_current_menu_index = OPENGL_MENU;
		s_ref_list[s_current_menu_index].curvalue = REF_OPENGL;
    s_ref_list[0].curvalue=0;
    s_ref_list[2].curvalue=2;
  }
  else
  {
    s_current_menu_index = 2;
    s_ref_list[s_current_menu_index].curvalue = 2;
  	s_ref_list[0].curvalue=0;
    s_ref_list[1].curvalue=1;
  }
	s_software_menu.x = viddef.width * 0.50;
	s_software_menu.nitems = 0;
	s_opengl_menu.x = viddef.width * 0.50;
	s_opengl_menu.nitems = 0;
  s_openglnolru_menu.x = viddef.width * 0.50;
  s_openglnolru_menu.nitems = 0;

	for ( i = 0; i < 2; i++ )
	{
		s_ref_list[i].generic.type = MTYPE_SPINCONTROL;
		s_ref_list[i].generic.name = "driver";
		s_ref_list[i].generic.x = 0;
		s_ref_list[i].generic.y = 0;
		s_ref_list[i].generic.callback = DriverCallback;
		s_ref_list[i].itemnames = refs;

		s_mode_list[i].generic.type = MTYPE_SPINCONTROL;
		s_mode_list[i].generic.name = "video mode";
		s_mode_list[i].generic.x = 0;
		s_mode_list[i].generic.y = 10;
		s_mode_list[i].itemnames = resolutions;

		s_screensize_slider[i].generic.type     = MTYPE_SLIDER;
		s_screensize_slider[i].generic.x                = 0;
		s_screensize_slider[i].generic.y                = 20;
		s_screensize_slider[i].generic.name     = "screen size";
		s_screensize_slider[i].minvalue = 3;
		s_screensize_slider[i].maxvalue = 12;
		s_screensize_slider[i].generic.callback = ScreenSizeCallback;

		s_brightness_slider[i].generic.type     = MTYPE_SLIDER;
		s_brightness_slider[i].generic.x        = 0;
		s_brightness_slider[i].generic.y        = 30;
		s_brightness_slider[i].generic.name     = "brightness";
		s_brightness_slider[i].generic.callback = BrightnessCallback;
		s_brightness_slider[i].minvalue = 5;
		s_brightness_slider[i].maxvalue = 13;
		s_brightness_slider[i].curvalue = ( 1.3 - vid_gamma->value + 0.5 ) * 10;

		s_fs_box[i].generic.type = MTYPE_SPINCONTROL;
		s_fs_box[i].generic.x   = 0;
		s_fs_box[i].generic.y   = 40;
		s_fs_box[i].generic.name        = "fullscreen";
		s_fs_box[i].itemnames = yesno_names;
		s_fs_box[i].curvalue = vid_fullscreen->value;

		s_defaults_action[i].generic.type = MTYPE_ACTION;
		s_defaults_action[i].generic.name = "reset to defaults";
		s_defaults_action[i].generic.x    = 0;
		s_defaults_action[i].generic.y    = 120;
		s_defaults_action[i].generic.callback = ResetDefaults;

		s_cancel_action[i].generic.type = MTYPE_ACTION;
		s_cancel_action[i].generic.name = "cancel";
		s_cancel_action[i].generic.x    = 0;
		s_cancel_action[i].generic.y    = 130;
		s_cancel_action[i].generic.callback = CancelChanges;
	}

	s_stipple_box.generic.type = MTYPE_SPINCONTROL;
	s_stipple_box.generic.x = 0;
	s_stipple_box.generic.y = 60;
	s_stipple_box.generic.name      = "stipple alpha";
	s_stipple_box.curvalue = sw_stipplealpha->value;
	s_stipple_box.itemnames = yesno_names;

	s_tq_slider.generic.type        = MTYPE_SLIDER;
	s_tq_slider.generic.x           = 0;
	s_tq_slider.generic.y           = 60;
	s_tq_slider.generic.name        = "texture quality";
	s_tq_slider.minvalue = 0;
	s_tq_slider.maxvalue = 3;
	s_tq_slider.curvalue = 3-gl_picmip->value;

	s_paletted_texture_box.generic.type = MTYPE_SPINCONTROL;
	s_paletted_texture_box.generic.x        = 0;
	s_paletted_texture_box.generic.y        = 70;
	s_paletted_texture_box.generic.name     = "8-bit textures";
	s_paletted_texture_box.itemnames = yesno_names;
	s_paletted_texture_box.curvalue = gl_ext_palettedtexture->value;

	s_finish_box.generic.type = MTYPE_SPINCONTROL;
	s_finish_box.generic.x  = 0;
	s_finish_box.generic.y  = 80;
	s_finish_box.generic.name       = "sync every frame";
	s_finish_box.curvalue = gl_finish->value;
	s_finish_box.itemnames = yesno_names;

	s_lockmode_box.generic.type = MTYPE_SPINCONTROL;
	s_lockmode_box.generic.x = 0;
	s_lockmode_box.generic.y = 90;
	s_lockmode_box.generic.name ="lockmode";
	if( strcmp( gl_lockmode->string, "SMART" ) == 0 )
		s_lockmode_box.curvalue = 1;
	else
		s_lockmode_box.curvalue = 0;
	s_lockmode_box.itemnames = lockmode_names;

	Menu_AddItem( &s_software_menu, ( void * ) &s_ref_list[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_mode_list[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_screensize_slider[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_brightness_slider[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_fs_box[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_stipple_box );

	Menu_AddItem( &s_opengl_menu, ( void * ) &s_ref_list[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_mode_list[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_screensize_slider[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_brightness_slider[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_fs_box[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_tq_slider );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_paletted_texture_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_finish_box );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_lockmode_box );

	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_ref_list[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_mode_list[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_screensize_slider[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_brightness_slider[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_fs_box[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_tq_slider );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_paletted_texture_box );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_finish_box );

	Menu_AddItem( &s_software_menu, ( void * ) &s_defaults_action[SOFTWARE_MENU] );
	Menu_AddItem( &s_software_menu, ( void * ) &s_cancel_action[SOFTWARE_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_defaults_action[OPENGL_MENU] );
	Menu_AddItem( &s_opengl_menu, ( void * ) &s_cancel_action[OPENGL_MENU] );

	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_defaults_action[OPENGLNOLRU_MENU] );
	Menu_AddItem( &s_openglnolru_menu, ( void * ) &s_cancel_action[OPENGLNOLRU_MENU] );

	Menu_Center( &s_software_menu );
	Menu_Center( &s_opengl_menu );
  Menu_Center( &s_openglnolru_menu);
	s_opengl_menu.x -= 8;
  s_openglnolru_menu.x -= 8;
	s_software_menu.x -= 8;
}

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	int w, h;

	if ( s_current_menu_index == 0 )
		s_current_menu = &s_software_menu;
	else //if (s_current_menu_index == 1)
		s_current_menu = &s_opengl_menu;
//  else s_current_menu = &s_openglnolru_menu;

	/*
	** draw the banner
	*/
	re.DrawGetPicSize( &w, &h, "m_banner_video" );
	re.DrawPic( viddef.width / 2 - w / 2, viddef.height /2 - 110, "m_banner_video" );

	/*
	** move cursor to a reasonable starting position
	*/
	Menu_AdjustCursor( s_current_menu, 1 );

	/*
	** draw the menu
	*/
	Menu_Draw( s_current_menu );
}

/*
================
VID_MenuKey
================
*/
const char *VID_MenuKey( int key )
{
	menuframework_s *m = s_current_menu;
	static const char *sound = "misc/menu1.wav";

	switch ( key )
	{
	case K_ESCAPE:
		ApplyChanges( 0 );
		return NULL;
	case K_KP_UPARROW:
	case K_UPARROW:
		m->cursor--;
		Menu_AdjustCursor( m, -1 );
		break;
	case K_KP_DOWNARROW:
	case K_DOWNARROW:
		m->cursor++;
		Menu_AdjustCursor( m, 1 );
		break;
	case K_KP_LEFTARROW:
	case K_LEFTARROW:
		Menu_SlideItem( m, -1 );
		break;
	case K_KP_RIGHTARROW:
	case K_RIGHTARROW:
		Menu_SlideItem( m, 1 );
		break;
	case K_KP_ENTER:
	case K_ENTER:
		if ( !Menu_SelectItem( m ) )
			ApplyChanges( NULL );
		break;
	}

	return sound;
}


